r"""
Group algebras of root lattice realizations
"""
# ****************************************************************************
#       Copyright (C) 2013 Nicolas M. Thiery <nthiery at users.sf.net>
#                          Anne Schilling <anne at math.ucdavis.edu>
#                          Mark Shimozono <mshimo at vt.edu>
#                          Daniel Bump
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import functools
import operator
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import lazy_import
from sage.misc.misc_c import prod
from sage.categories.algebra_functor import AlgebrasCategory
lazy_import('sage.rings.integer_ring', 'ZZ')
from sage.modules.free_module_element import vector
from sage.combinat.root_system.hecke_algebra_representation import HeckeAlgebraRepresentation


class Algebras(AlgebrasCategory):
    """
    The category of group algebras of root lattice realizations.

    This includes typically weight rings (group algebras of weight lattices).

    TESTS::

        sage: for ct in CartanType.samples(crystallographic=True): # long time
        ....:     TestSuite(RootSystem(ct).root_lattice().algebra(QQ)).run()
    """

    class ParentMethods:

        def _repr_(self):
            r"""
            EXAMPLES::

                sage: RootSystem(["A",2,1]).ambient_space().algebra(QQ) # indirect doctest
                Algebra of the Ambient space of the Root system of type ['A', 2, 1] over Rational Field
            """
            return "Algebra of the %s over %s" % (self.basis().keys(), self.base_ring())

        def some_elements(self):
            r"""
            Return some elements of the algebra ``self``.

            EXAMPLES::

                sage: A = RootSystem(["A",2,1]).ambient_space().algebra(QQ)
                sage: A.some_elements()
                [B[2*e[0] + 2*e[1] + 3*e[2]]...]
                sage: A.some_elements()                                                 # needs sage.graphs
                [B[2*e[0] + 2*e[1] + 3*e[2]],
                 B[-e[0] + e[2] + e['delta']],
                 B[e[0] - e[1]],
                 B[e[1] - e[2]],
                 B[e['deltacheck']],
                 B[e[0] + e['deltacheck']],
                 B[e[0] + e[1] + e['deltacheck']]]

                sage: A = RootSystem(["B",2]).weight_space().algebra(QQ)
                sage: A.some_elements()
                [B[2*Lambda[1] + 2*Lambda[2]], ... B[Lambda[1]], B[Lambda[2]]]
                sage: A.some_elements()                                                 # needs sage.graphs
                [B[2*Lambda[1] + 2*Lambda[2]],
                 B[2*Lambda[1] - 2*Lambda[2]],
                 B[-Lambda[1] + 2*Lambda[2]],
                 B[Lambda[1]],
                 B[Lambda[2]]]
            """
            return [self.monomial(weight) for weight in self.basis().keys().some_elements()]

        @cached_method
        def cartan_type(self):
            r"""
            Return the Cartan type of ``self``.

            EXAMPLES::

                sage: A = RootSystem(["A",2,1]).ambient_space().algebra(QQ)
                sage: A.cartan_type()
                ['A', 2, 1]
                sage: A = RootSystem(["B",2]).weight_space().algebra(QQ)
                sage: A.cartan_type()
                ['B', 2]
            """
            return self.basis().keys().cartan_type()

        def from_polynomial(self, p):
            """
            Construct an element of ``self`` from a polynomial `p`.

            INPUT:

            - ``p`` -- a polynomial

            EXAMPLES::

                sage: L = RootSystem(["A",2]).ambient_lattice()
                sage: KL = L.algebra(QQ)
                sage: x,y,z = QQ['x,y,z'].gens()
                sage: KL.from_polynomial(x)
                B[(1, 0, 0)]
                sage: KL.from_polynomial(x^2*y + 2*y - z)
                B[(2, 1, 0)] + 2*B[(0, 1, 0)] - B[(0, 0, 1)]

            TESTS::

                sage: KL.from_polynomial(x).leading_support().parent() is L
                True
                sage: KL.from_polynomial(x-x)
                0
                sage: KL.from_polynomial(x-x).parent() is KL
                True

            .. TODO:: make this work for Laurent polynomials too
            """
            L = self.basis().keys()
            return self.sum_of_terms((L.from_vector(vector(t)), c)
                                     for t, c in p.monomial_coefficients().items())

        @cached_method
        def divided_difference_on_basis(self, weight, i):
            r"""
            Return the result of applying the `i`-th divided difference on ``weight``.

            INPUT:

            - ``weight`` -- a weight
            - ``i`` -- an element of the index set

            .. TODO:: type free definition (Viviane's definition uses that we are in the ambient space)

            EXAMPLES::

                sage: L = RootSystem(["A",1]).ambient_space()
                sage: KL = L.algebra(QQ)
                sage: KL.divided_difference_on_basis(L((2,2)), 1) # todo: not implemented
                0
                sage: KL.divided_difference_on_basis(L((3,0)), 1) # todo: not implemented
                B[(2, 0)] + B[(1, 1)] + B[(0, 2)]
                sage: KL.divided_difference_on_basis(L((0,3)), 1) # todo: not implemented
                -B[(2, 0)] - B[(1, 1)] - B[(0, 2)]

            In type `A` and in the ambient lattice, we recover the
            usual action of divided differences polynomials::

                sage: x,y = QQ['x,y'].gens()
                sage: d = lambda p: (p - p(y,x)) / (x-y)
                sage: d(x^2*y^2)
                0
                sage: d(x^3)
                x^2 + x*y + y^2
                sage: d(y^3)
                -x^2 - x*y - y^2
            """
            raise NotImplementedError()

        @cached_method
        def isobaric_divided_difference_on_basis(self, weight, i):
            r"""
            Return the result of applying the `i`-th isobaric divided difference on ``weight``.

            INPUT:

            - ``weight`` -- a weight
            - ``i`` -- an element of the index set

            .. SEEALSO:: :meth:`demazure_operators`

            EXAMPLES::

                sage: L = RootSystem(["A",1]).ambient_space()
                sage: KL = L.algebra(QQ)
                sage: KL.isobaric_divided_difference_on_basis(L((2,2)), 1)
                B[(2, 2)]
                sage: KL.isobaric_divided_difference_on_basis(L((3,0)), 1)
                B[(1, 2)] + B[(2, 1)] + B[(3, 0)] + B[(0, 3)]
                sage: KL.isobaric_divided_difference_on_basis(L((0,3)), 1)
                -B[(1, 2)] - B[(2, 1)]

            In type `A` and in the ambient lattice, we recover the
            usual action of divided differences on polynomials::

                sage: x,y = QQ['x,y'].gens()
                sage: d = lambda p: (x*p - (x*p)(y,x)) / (x-y)
                sage: d(x^2*y^2)
                x^2*y^2
                sage: d(x^3)
                x^3 + x^2*y + x*y^2 + y^3
                sage: d(y^3)
                -x^2*y - x*y^2

            REFERENCES:

            .. [Lascoux2003] Alain Lascoux, Symmetric functions and combinatorial operators on polynomials,
               CBMS Regional Conference Series in Mathematics, 99, 2003.
            """
            P = self.basis().keys()  # the root lattice realization
            n = weight.scalar(P.simple_coroot(i))
            if n not in ZZ:
                raise ValueError("the weight does not have an integral scalar product with the coroot")
            alphai = P.simple_root(i)
            if n >= 0:
                return self.sum_of_monomials(weight-j*alphai for j in range(n + 1))
            return -self.sum_of_monomials(weight-j*alphai for j in range(n + 1, 0))

        def demazure_operators(self):
            r"""
            Return the Demazure operators acting on ``self``.

            The `i`-th Demazure operator is defined by:

            .. MATH::

                \pi_i = \frac{ 1 - e^{-\alpha_i}s_i }{ 1-e^{-\alpha_i} }

            It acts on `e^\lambda`, for `\lambda` a weight, by:

            .. MATH::

                \pi_i e^\lambda = \frac{e^\lambda - e^{-\alpha_i+s_i\lambda}}{1-e^{-\alpha_i}}

            This matches with Lascoux' definition [Lascoux2003]_ of `\pi_i`, and
            with the `i`-th Demazure operator of [Kumar1987]_, which also works for
            general Kac-Moody types.

            REFERENCES:

            .. [Kumar1987] \S. Kumar, Demazure character formula in arbitrary Kac-Moody setting,
               Invent. Math. 89 (1987), no. 2, 395-423.

            EXAMPLES:

            We compute some Schur functions, as images of dominant
            monomials under the action of the maximal isobaric divided
            difference `\Delta_{w_0}`::

                sage: L = RootSystem(["A",2]).ambient_lattice()
                sage: KL = L.algebra(QQ)
                sage: w0 = tuple(L.weyl_group().long_element().reduced_word())          # needs sage.libs.gap
                sage: pi = KL.demazure_operators()
                sage: pi0 = pi[w0]                                                      # needs sage.libs.gap
                sage: pi0(KL.monomial(L((2,1))))                                        # needs sage.libs.gap
                2*B[(1, 1, 1)] + B[(1, 2, 0)] + B[(1, 0, 2)] + B[(2, 1, 0)]
                 + B[(2, 0, 1)] + B[(0, 1, 2)] + B[(0, 2, 1)]

            Let us make the result into an actual polynomial::

                sage: P = QQ['x,y,z']
                sage: pi0(KL.monomial(L((2,1)))).expand(P.gens())                       # needs sage.libs.gap
                x^2*y + x*y^2 + x^2*z + 2*x*y*z + y^2*z + x*z^2 + y*z^2

            This is indeed a Schur function::

                sage: s = SymmetricFunctions(QQ).s()                                    # needs sage.combinat
                sage: s[2,1].expand(3, P.variable_names())                              # needs sage.combinat
                x^2*y + x*y^2 + x^2*z + 2*x*y*z + y^2*z + x*z^2 + y*z^2

            Let us check this systematically on Schur functions of degree 6::

                sage: for p in Partitions(6, max_length=3).list():                      # needs sage.combinat sage.libs.gap
                ....:     assert (s.monomial(p).expand(3, P.variable_names())
                ....:              == pi0(KL.monomial(L(tuple(p)))).expand(P.gens()))

            We check systematically that these operators satisfy the Iwahori-Hecke algebra relations::

                sage: for cartan_type in CartanType.samples(crystallographic=True):  # long time (12s)
                ....:     L = RootSystem(cartan_type).weight_lattice()
                ....:     KL = L.algebra(QQ)
                ....:     T = KL.demazure_operators()
                ....:     T._test_relations()

                sage: L = RootSystem(['A',1,1]).weight_lattice()
                sage: KL = L.algebra(QQ)
                sage: T = KL.demazure_operators()
                sage: T._test_relations()

            .. WARNING::

                The Demazure operators are only defined if all the
                elements in the support have integral scalar products
                with the coroots (basically, they are in the weight
                lattice). Otherwise an error is raised::

                    sage: L = RootSystem(CartanType(["G",2]).dual()).ambient_space()
                    sage: KL = L.algebra(QQ)
                    sage: pi = KL.demazure_operators()
                    sage: pi[1](KL.monomial(L([0,0,1])))
                    Traceback (most recent call last):
                    ...
                    ValueError: the weight does not have an integral scalar product with the coroot
            """
            return HeckeAlgebraRepresentation(self, self.isobaric_divided_difference_on_basis, self.cartan_type(), 0, 1, side='left')

        def _test_demazure_operators(self, **options):
            """
            Test that the Demazure operators satisfy their defining formulas.

            EXAMPLES::

                sage: RootSystem(["A",2]).root_lattice().algebra(QQ)._test_demazure_operators()
            """
            tester = self._tester(**options)
            try:
                pi = self.demazure_operators()
                L = self.basis().keys()
                alpha = L.simple_roots()
                alphacheck = L.simple_coroots()
                s = L.simple_reflections()
                for i in self.cartan_type().index_set():
                    emalphai = self.monomial(-alpha[i])  # X^{-\alpha_i}
                    for weight in L.some_elements():
                        if weight.scalar(alphacheck[i]) not in ZZ:
                            # Demazure operators are not defined in this case
                            continue
                        x = self.monomial(weight)
                        result = pi[i](x)
                        tester.assertEqual(result * (self.one() - emalphai),
                                           x - emalphai * x.map_support(s[i]))
            except ImportError:
                pass

        def demazure_lusztig_operator_on_basis(self, weight, i, q1, q2, convention='antidominant'):
            r"""
            Return the result of applying the `i`-th Demazure-Lusztig operator on ``weight``.

            INPUT:

            - ``weight`` -- an element `\lambda` of the weight lattice
            - ``i`` -- an element of the index set
            - ``q1``, ``q2`` -- two elements of the ground ring
            - ``convention`` -- ``'antidominant'``, ``'bar'``, or ``'dominant'`` (default: ``'antidominant'``)

            See :meth:`demazure_lusztig_operators` for the details.

            EXAMPLES::

                sage: L = RootSystem(["A",1]).ambient_space()
                sage: K = QQ['q1,q2']
                sage: q1, q2 = K.gens()
                sage: KL = L.algebra(K)
                sage: KL.demazure_lusztig_operator_on_basis(L((2,2)), 1, q1, q2)
                q1*B[(2, 2)]
                sage: KL.demazure_lusztig_operator_on_basis(L((3,0)), 1, q1, q2)
                (q1+q2)*B[(1, 2)] + (q1+q2)*B[(2, 1)] + (q1+q2)*B[(3, 0)] + q1*B[(0, 3)]
                sage: KL.demazure_lusztig_operator_on_basis(L((0,3)), 1, q1, q2)
                -(q1+q2)*B[(1, 2)] - (q1+q2)*B[(2, 1)] - q2*B[(3, 0)]

            At `q_1=1` and `q_2=0` we recover the action of the isobaric divided differences `\pi_i`::

                sage: KL.demazure_lusztig_operator_on_basis(L((2,2)), 1, 1, 0)
                B[(2, 2)]
                sage: KL.demazure_lusztig_operator_on_basis(L((3,0)), 1, 1, 0)
                B[(1, 2)] + B[(2, 1)] + B[(3, 0)] + B[(0, 3)]
                sage: KL.demazure_lusztig_operator_on_basis(L((0,3)), 1, 1, 0)
                -B[(1, 2)] - B[(2, 1)]

            Or `1-\pi_i` for ``bar=True``::

                sage: KL.demazure_lusztig_operator_on_basis(L((2,2)), 1, 1, 0, convention='bar')
                0
                sage: KL.demazure_lusztig_operator_on_basis(L((3,0)), 1, 1, 0, convention='bar')
                -B[(1, 2)] - B[(2, 1)] - B[(0, 3)]
                sage: KL.demazure_lusztig_operator_on_basis(L((0,3)), 1, 1, 0, convention='bar')
                B[(1, 2)] + B[(2, 1)] + B[(0, 3)]

            At `q_1=1` and `q_2=-1` we recover the action of the simple reflection `s_i`::

                sage: KL.demazure_lusztig_operator_on_basis(L((2,2)), 1, 1, -1)
                B[(2, 2)]
                sage: KL.demazure_lusztig_operator_on_basis(L((3,0)), 1, 1, -1)
                B[(0, 3)]
                sage: KL.demazure_lusztig_operator_on_basis(L((0,3)), 1, 1, -1)
                B[(3, 0)]
            """
            if convention == "dominant":
                weight = -weight
            pi_on_weight = self.isobaric_divided_difference_on_basis(weight, i)
            if convention == "bar":
                pi_on_weight = self.monomial(weight) - pi_on_weight
            result = (q1+q2) * pi_on_weight - self.term(weight.simple_reflection(i), q2)
            if convention == "dominant":
                return result.map_support(operator.neg)
            return result

        def demazure_lusztig_operators(self, q1, q2, convention='antidominant'):
            r"""
            Return the Demazure-Lusztig operators acting on ``self``.

            INPUT:

            - ``q1``, ``q2`` -- two elements of the ground ring
            - ``convention`` -- "antidominant", "bar", or "dominant" (default: ``'antidominant'``)

            If `R` is the parent weight ring, the Demazure-Lusztig
            operator `T_i` is the linear map `R\rightarrow R` obtained
            by interpolating between the isobaric divided difference
            operator `\pi_i` (see :meth:`.isobaric_divided_difference_on_basis`)
            and the simple reflection `s_i`.

            .. MATH::

                (q_1+q_2) \pi_i - q_2 s_i

            The Demazure-Lusztig operators give the usual
            representation of the operator `T_i` of the (affine) Hecke
            algebra with eigenvalues `q_1` and `q_2` associated to the
            Weyl group.

            Several variants are available to match with various
            conventions used in the literature:

            - "bar" replaces `\pi_i` in the formula above by
              `\overline{\pi}_i = (1-\pi_i)`.
            - "dominant" conjugates the operator by
              `x^\lambda \mapsto x^-\lambda`.

            The names dominant and antidominant for the conventions were chosen with regards to
            the nonsymmetric Macdonald polynomials. The `Y` operators for the Macdonald polynomials
            in the "dominant" convention satisfy `Y_\lambda = T_{t_{\lambda}}` for `\lambda` dominant.
            This is also the convention used in [Haiman06]_. For the "antidominant" convention,
            `Y_\lambda = T_{t_{\lambda}}` with `\lambda` antidominant.

            .. SEEALSO::

                - :meth:`demazure_lusztig_operator_on_basis`.
                - :class:`~.non_symmetric_macdonald_polynomials.NonSymmetricMacdonaldPolynomials`.

            REFERENCES:

            .. [Lusztig1985] \G. Lusztig,
               *Equivariant K-theory and representations of Hecke algebras*,
               Proc. Amer. Math. Soc. 94 (1985), no. 2, 337-342.

            .. [Cherednik1995] \I. Cherednik,
               *Nonsymmetric Macdonald polynomials*. IMRN 10, 483-515 (1995).

            EXAMPLES::

                sage: L = RootSystem(["A",1]).ambient_space()
                sage: K = QQ['q1,q2'].fraction_field()
                sage: q1, q2 = K.gens()
                sage: KL = L.algebra(K)
                sage: T = KL.demazure_lusztig_operators(q1, q2)
                sage: Tbar = KL.demazure_lusztig_operators(q1, q2, convention='bar')
                sage: Tdominant = KL.demazure_lusztig_operators(q1, q2,
                ....:                                           convention='dominant')
                sage: x = KL.monomial(L((3,0)))
                sage: T[1](x)
                (q1+q2)*B[(1, 2)] + (q1+q2)*B[(2, 1)] + (q1+q2)*B[(3, 0)] + q1*B[(0, 3)]
                sage: Tbar[1](x)
                -(q1+q2)*B[(1, 2)] - (q1+q2)*B[(2, 1)] - (q1+2*q2)*B[(0, 3)]
                sage: Tbar[1](x) + T[1](x)
                (q1+q2)*B[(3, 0)] - 2*q2*B[(0, 3)]
                sage: Tdominant[1](x)
                -(q1+q2)*B[(1, 2)] - (q1+q2)*B[(2, 1)] - q2*B[(0, 3)]

                sage: Tdominant.Tw_inverse(1)(KL.monomial(-L.simple_root(1)))
                -((q1+q2)/(q1*q2))*B[(0, 0)] - 1/q2*B[(1, -1)]

            We repeat similar computation in the affine setting::

                sage: L = RootSystem(["A",2,1]).ambient_space()
                sage: K = QQ['q1,q2'].fraction_field()
                sage: q1, q2 = K.gens()
                sage: KL = L.algebra(K)
                sage: T = KL.demazure_lusztig_operators(q1, q2)
                sage: Tbar = KL.demazure_lusztig_operators(q1, q2, convention='bar')
                sage: Tdominant = KL.demazure_lusztig_operators(q1, q2,
                ....:                                           convention='dominant')
                sage: e = L.basis()
                sage: x = KL.monomial(3*e[0])
                sage: T[1](x)
                (q1+q2)*B[e[0] + 2*e[1]] + (q1+q2)*B[2*e[0] + e[1]]
                 + (q1+q2)*B[3*e[0]] + q1*B[3*e[1]]
                sage: Tbar[1](x)
                -(q1+q2)*B[e[0] + 2*e[1]] - (q1+q2)*B[2*e[0] + e[1]] - (q1+2*q2)*B[3*e[1]]
                sage: Tbar[1](x) + T[1](x)
                (q1+q2)*B[3*e[0]] - 2*q2*B[3*e[1]]
                sage: Tdominant[1](x)
                -(q1+q2)*B[e[0] + 2*e[1]] - (q1+q2)*B[2*e[0] + e[1]] - q2*B[3*e[1]]
                sage: Tdominant.Tw_inverse(1)(KL.monomial(-L.simple_root(1)))
                -((q1+q2)/(q1*q2))*B[0] - 1/q2*B[e[0] - e[1]]

            One can obtain iterated operators by passing a reduced
            word or an element of the Weyl group::

                sage: T[1,2](x)
                (q1^2+2*q1*q2+q2^2)*B[e[0] + e[1] + e[2]] +
                (q1^2+2*q1*q2+q2^2)*B[e[0] + 2*e[1]] +
                (q1^2+q1*q2)*B[e[0] + 2*e[2]] + (q1^2+2*q1*q2+q2^2)*B[2*e[0] + e[1]] +
                (q1^2+q1*q2)*B[2*e[0] + e[2]] + (q1^2+q1*q2)*B[3*e[0]] +
                (q1^2+q1*q2)*B[e[1] + 2*e[2]] + (q1^2+q1*q2)*B[2*e[1] + e[2]] +
                (q1^2+q1*q2)*B[3*e[1]] + q1^2*B[3*e[2]]

            and use that to check, for example, the braid relations::

                sage: T[1,2,1](x) - T[2,1,2](x)
                0

            The operators satisfy the relations of the affine Hecke
            algebra with parameters `q_1`, `q_2`::

                sage: T._test_relations()
                sage: Tdominant._test_relations()
                sage: Tbar._test_relations() #-q2,q1+2*q2   # todo: not implemented: set the appropriate eigenvalues!

            And the `\bar{T}` are basically the inverses of the `T` s::

                sage: Tinv = KL.demazure_lusztig_operators(2/q1 + 1/q2, -1/q1,
                ....:                                      convention='bar')
                sage: [Tinv[1](T[1](x)) - x for x in KL.some_elements()]                # needs sage.graphs
                [0, 0, 0, 0, 0, 0, 0]

            We check that `\Lambda_1-\Lambda_0` is an eigenvector for
            the `Y` s in affine type::

                sage: K = QQ['q,q1,q2'].fraction_field()
                sage: q,q1,q2 = K.gens()
                sage: L = RootSystem(["A",2,1]).ambient_space()
                sage: L0 = L.classical()
                sage: Lambda = L.fundamental_weights()                                  # needs sage.graphs
                sage: alphacheck = L0.simple_coroots()
                sage: KL = L.algebra(K)
                sage: T = KL.demazure_lusztig_operators(q1, q2, convention='dominant')
                sage: Y = T.Y()
                sage: alphacheck = Y.keys().alpha()  # alpha of coroot lattice is alphacheck
                sage: alphacheck
                Finite family {0: alphacheck[0], 1: alphacheck[1], 2: alphacheck[2]}
                sage: x = KL.monomial(Lambda[1] - Lambda[0]); x                         # needs sage.graphs
                B[e[0]]

            In fact it is not exactly an eigenvector, but the extra
            '\delta` term is to be interpreted as a `q` parameter::

                sage: # needs sage.graphs
                sage: Y[alphacheck[0]](KL.one())
                q2^2/q1^2*B[0]
                sage: Y[alphacheck[1]](x)
                -(q2^2/(-q1^2))*B[e[0] - e['delta']]
                sage: Y[alphacheck[2]](x)
                (q1/(-q2))*B[e[0]]
                sage: KL.q_project(Y[alphacheck[1]](x),q)
                -(q2^2/(-q*q1^2))*B[(1, 0, 0)]

                sage: # needs sage.graphs
                sage: KL.q_project(x, q)
                B[(1, 0, 0)]
                sage: KL.q_project(Y[alphacheck[0]](x),q)
                -q*q1/q2*B[(1, 0, 0)]
                sage: KL.q_project(Y[alphacheck[1]](x),q)
                -(q2^2/(-q*q1^2))*B[(1, 0, 0)]
                sage: KL.q_project(Y[alphacheck[2]](x),q)
                (q1/(-q2))*B[(1, 0, 0)]

            We now check systematically that the Demazure-Lusztig
            operators satisfy the relations of the Iwahori-Hecke
            algebra::

                sage: K = QQ['q1,q2']
                sage: q1, q2 = K.gens()
                sage: for cartan_type in CartanType.samples(crystallographic=True):  # long time (12s)
                ....:    L = RootSystem(cartan_type).root_lattice()
                ....:    KL = L.algebra(K)
                ....:    T = KL.demazure_lusztig_operators(q1,q2)
                ....:    T._test_relations()

                sage: for cartan_type in CartanType.samples(crystallographic=True):  # long time (12s)
                ....:    L = RootSystem(cartan_type).weight_lattice()
                ....:    KL = L.algebra(K)
                ....:    T = KL.demazure_lusztig_operators(q1,q2)
                ....:    T._test_relations()

            Recall that the Demazure-Lusztig operators are only
            defined when all monomials belong to the weight lattice.
            Thus, in the group algebra of the ambient space, we need
            to specify explicitly the elements on which to run the
            tests::

                sage: for cartan_type in CartanType.samples(crystallographic=True):  # long time (12s)
                ....:    L = RootSystem(cartan_type).ambient_space()
                ....:    KL = L.algebra(K)
                ....:    weight_lattice = RootSystem(cartan_type).weight_lattice(extended=L.is_extended())
                ....:    elements = [ KL.monomial(L(weight)) for weight in weight_lattice.some_elements() ]
                ....:    T = KL.demazure_lusztig_operators(q1,q2)
                ....:    T._test_relations(elements=elements)
            """
            T_on_basis = functools.partial(self.demazure_lusztig_operator_on_basis,
                                           q1=q1, q2=q2, convention=convention)
            return HeckeAlgebraRepresentation(self, T_on_basis, self.cartan_type(), q1, q2, side='left')

        def demazure_lusztig_operator_on_classical_on_basis(self, weight, i, q, q1, q2, convention='antidominant'):
            r"""
            Return the result of applying the `i`-th Demazure-Lusztig operator on the classical weight ``weight`` embedded at level 0.

            INPUT:

            - ``weight`` -- a classical weight `\lambda`
            - ``i`` -- an element of the index set
            - ``q1``, ``q2`` -- two elements of the ground ring
            - ``convention`` -- ``'antidominant'``, ``'bar'``, or ``'dominant'`` (default: ``'antidominant'``)

            See :meth:`demazure_lusztig_operators` for the details.

            .. TODO::

                - Optimize the code to only do the embedding/projection for T_0
                - Add an option to specify at which level one wants to
                  work. Currently this is level 0.

            EXAMPLES::

                sage: L = RootSystem(["A",1,1]).ambient_space()
                sage: L0 = L.classical()
                sage: K = QQ['q,q1,q2']
                sage: q, q1, q2 = K.gens()
                sage: KL = L.algebra(K)
                sage: KL0 = L0.algebra(K)

            These operators coincide with the usual Demazure-Lusztig
            operators::

                sage: KL.demazure_lusztig_operator_on_classical_on_basis(L0((2,2)),     # needs sage.graphs
                ....:                                                    1, q, q1, q2)
                q1*B[(2, 2)]
                sage: KL0.demazure_lusztig_operator_on_basis(L0((2,2)), 1, q1, q2)
                q1*B[(2, 2)]

                sage: KL.demazure_lusztig_operator_on_classical_on_basis(L0((3,0)),     # needs sage.graphs
                ....:                                                    1, q, q1, q2)
                (q1+q2)*B[(1, 2)] + (q1+q2)*B[(2, 1)] + (q1+q2)*B[(3, 0)] + q1*B[(0, 3)]
                sage: KL0.demazure_lusztig_operator_on_basis(L0((3,0)), 1, q1, q2)
                (q1+q2)*B[(1, 2)] + (q1+q2)*B[(2, 1)] + (q1+q2)*B[(3, 0)] + q1*B[(0, 3)]

            except that we now have an action of `T_0`, which introduces some `q` s::

                sage: KL.demazure_lusztig_operator_on_classical_on_basis(L0((2,2)),     # needs sage.graphs
                ....:                                                    0, q, q1, q2)
                q1*B[(2, 2)]
                sage: KL.demazure_lusztig_operator_on_classical_on_basis(L0((3,0)),     # needs sage.graphs
                ....:                                                    0, q, q1, q2)
                -(q^2*q1+q^2*q2)*B[(1, 2)] - (q*q1+q*q2)*B[(2, 1)] - q^3*q2*B[(0, 3)]
            """
            L = self.basis().keys()
            weight = L.embed_at_level(weight, 0)
            return self.q_project(self.demazure_lusztig_operator_on_basis(weight, i, q1, q2, convention=convention), q)

        def demazure_lusztig_operators_on_classical(self, q, q1, q2, convention='antidominant'):
            r"""
            Return the Demazure-Lusztig operators acting at level 1 on ``self.classical()``.

            INPUT:

            - ``q``, ``q1``, ``q2`` -- three elements of the ground ring
            - ``convention`` -- ``'antidominant'``, ``'bar'``, or ``'dominant'`` (default: ``'antidominant'``)

            Let `KL` be the group algebra of an affine weight lattice
            realization `L`. The Demazure-Lusztig operators for `KL`
            act on the group algebra of the corresponding classical
            weight lattice by embedding it at level 1, and projecting
            back.

            .. SEEALSO::

                - :meth:`demazure_lusztig_operators`.
                - :meth:`demazure_lusztig_operator_on_classical_on_basis`.
                - :meth:`q_project`

            EXAMPLES::

                sage: L = RootSystem(["A",1,1]).ambient_space()
                sage: K = QQ['q,q1,q2'].fraction_field()
                sage: q, q1, q2 = K.gens()
                sage: KL = L.algebra(K)
                sage: KL0 = KL.classical()
                sage: L0 = KL0.basis().keys()
                sage: T = KL.demazure_lusztig_operators_on_classical(q, q1, q2)         # needs sage.graphs

                sage: x = KL0.monomial(L0((3,0))); x
                B[(3, 0)]

            For `T_1,\dots` we recover the usual Demazure-Lusztig operators::

                sage: T[1](x)                                                           # needs sage.graphs
                (q1+q2)*B[(1, 2)] + (q1+q2)*B[(2, 1)] + (q1+q2)*B[(3, 0)] + q1*B[(0, 3)]

            For `T_0`, we can note that, in the projection, `\delta`
            is mapped to `q`::

                sage: T[0](x)                                                           # needs sage.graphs
                -(q^2*q1+q^2*q2)*B[(1, 2)] - (q*q1+q*q2)*B[(2, 1)] - q^3*q2*B[(0, 3)]

            Note that there is no translation part, and in particular
            1 is an eigenvector for all `T_i`'s::

                sage: # needs sage.graphs
                sage: T[0](KL0.one())
                q1*B[(0, 0)]
                sage: T[1](KL0.one())
                q1*B[(0, 0)]
                sage: Y = T.Y()
                sage: alphacheck = Y.keys().simple_roots()
                sage: Y[alphacheck[0]](KL0.one())
                -q2/(q*q1)*B[(0, 0)]

            Matching with Ion Bogdan's hand calculations from 3/15/2013::

                sage: L = RootSystem(["A",1,1]).weight_space(extended=True)
                sage: K = QQ['q,u'].fraction_field()
                sage: q, u = K.gens()
                sage: KL = L.algebra(K)
                sage: KL0 = KL.classical()
                sage: L0 = KL0.basis().keys()
                sage: omega = L0.fundamental_weights()

                sage: # needs sage.graphs
                sage: T = KL.demazure_lusztig_operators_on_classical(q, u, -1/u,
                ....:                                                convention='dominant')
                sage: Y = T.Y()
                sage: alphacheck = Y.keys().simple_roots()
                sage: Ydelta = Y[Y.keys().null_root()]
                sage: Ydelta.word, Ydelta.signs, Ydelta.scalar
                ((), (), 1/q)
                sage: Y1 = Y[alphacheck[1]]
                sage: Y1.word, Y1.signs, Y1.scalar  # This is T_0 T_1 (T_1 acts first, then T_0); Ion gets T_1 T_0
                ((1, 0), (1, 1), 1)
                sage: Y0 = Y[alphacheck[0]]
                sage: Y0.word, Y0.signs, Y0.scalar  # This is 1/q T_1^-1 T_0^-1
                ((0, 1), (-1, -1), 1/q)

            Note that the following computations use the "dominant" convention::

                sage: # needs sage.graphs
                sage: T0 = T.Tw(0)
                sage: T0(KL0.monomial(omega[1]))
                q*u*B[-Lambda[1]] + ((u^2-1)/u)*B[Lambda[1]]
                sage: T0(KL0.monomial(2*omega[1]))
                ((q*u^2-q)/u)*B[0] + q^2*u*B[-2*Lambda[1]] + ((u^2-1)/u)*B[2*Lambda[1]]
                sage: T0(KL0.monomial(-omega[1]))
                1/(q*u)*B[Lambda[1]]
                sage: T0(KL0.monomial(-2*omega[1]))
                -((u^2-1)/(q*u))*B[0] + 1/(q^2*u)*B[2*Lambda[1]]
            """
            # In type BC dual we used q^2 and q elsewhere
            # Not sure this is the right thing to do or just a workaround ...
            # This probably makes up for the fact that, in type BC
            # dual, the null coroot is twice Sage's deltacheck
            # whereas the null root is delta. So we need to map delta
            # to q^2 in the q_projection.
            # Should this go in q_project instead?
            ct = self.cartan_type()
            a0check = ct.acheck()[ct.special_node()]
            T_on_basis = functools.partial(self.demazure_lusztig_operator_on_classical_on_basis,
                                           q1=q1, q2=q2, q=q**a0check, convention=convention)
            return HeckeAlgebraRepresentation(self.classical(), T_on_basis, self.cartan_type(), q1=q1, q2=q2, q=q, side='left')

        @cached_method
        def T0_check_on_basis(self, q1, q2, convention='antidominant'):
            r"""
            Return the `T_0^\vee` operator acting on the basis.

            This implements the formula for `T_{0'}` in Section 6.12 of [Haiman06]_.

            REFERENCES:

            .. [Haiman06] \M. Haiman, Cherednik algebras, Macdonald polynomials and combinatorics, ICM 2006.

            .. WARNING::

                The current implementation probably returns just
                nonsense, if the convention is not "dominant".

            EXAMPLES::

                sage: K = QQ['q1,q2'].fraction_field()
                sage: q1,q2 = K.gens()

                sage: L = RootSystem(["A",1,1]).ambient_space()
                sage: L0 = L.classical()
                sage: KL = L.algebra(K)
                sage: some_weights = L.fundamental_weights()                            # needs sage.graphs
                sage: f = KL.T0_check_on_basis(q1,q2, convention='dominant')            # needs sage.graphs
                sage: f(L0.zero())                                                      # needs sage.graphs
                (q1+q2)*B[(0, 0)] + q1*B[(1, -1)]

                sage: L = RootSystem(["A",3,1]).ambient_space()
                sage: L0 = L.classical()
                sage: KL = L.algebra(K)
                sage: some_weights = L0.fundamental_weights()
                sage: f = KL.T0_check_on_basis(q1,q2, convention='dominant')            # needs sage.graphs
                sage: f(L0.zero())       # not checked                                  # needs sage.graphs
                (q1+q2)*B[(0, 0, 0, 0)] + q1^3/q2^2*B[(1, 0, 0, -1)]

            The following results have not been checked::

                sage: for x in some_weights:                                            # needs sage.graphs
                ....:     print("{} : {}".format(x, f(x)))
                (1, 0, 0, 0) : q1*B[(1, 0, 0, 0)]
                (1, 1, 0, 0) : q1*B[(1, 1, 0, 0)]
                (1, 1, 1, 0) : q1*B[(1, 1, 1, 0)]

            Some examples for type `B_2^{(1)}` dual::

                sage: L = RootSystem("B2~*").ambient_space()
                sage: L0 = L.classical()
                sage: e = L.basis()
                sage: K = QQ['q,u'].fraction_field()
                sage: q,u = K.gens()
                sage: q1 = u
                sage: q2 = -1/u
                sage: KL = L.algebra(K)
                sage: KL0 = KL.classical()
                sage: f = KL.T0_check_on_basis(q1,q2, convention='dominant')            # needs sage.graphs
                sage: T = KL.twisted_demazure_lusztig_operators(q1,q2, convention='dominant')

            Direct calculation::

                sage: T.Tw(0)(KL0.monomial(L0([0,0])))                                  # needs sage.graphs
                ((u^2-1)/u)*B[(0, 0)] + u^3*B[(1, 1)]
                sage: KL.T0_check_on_basis(q1,q2, convention='dominant')(L0([0,0]))     # needs sage.graphs
                ((u^2-1)/u)*B[(0, 0)] + u^3*B[(1, 1)]

            Step by step calculation, comparing by hand with Mark Shimozono::

                sage: res = T.Tw(2)(KL0.monomial(L0([0,0]))); res
                u*B[(0, 0)]
                sage: res = res * KL0.monomial(L0([-1,1])); res
                u*B[(-1, 1)]
                sage: res = T.Tw_inverse(1)(res); res
                (u^2-1)*B[(0, 0)] + u^2*B[(1, -1)]
                sage: res = T.Tw_inverse(2)(res); res
                ((u^2-1)/u)*B[(0, 0)] + u^3*B[(1, 1)]
            """
            L = self.basis().keys()
            ct = L.cartan_type()
            special_node = ct.special_node()
            a0 = ct.a()[special_node]
            A0 = self.classical()
            T = A0.demazure_lusztig_operators(q1, q2, convention=convention)
            # TODO: use the formula expressing the inverse of T as a Demazure Lusztig operator? Or go through the affine action of T_0 for the dual
            L0 = A0.basis().keys()
            # The dominant short root of the classical system
            if ct.type() == 'BC':
                # CHECKME: this is not exactly phi, but phi rescaled
                # appropriately so that it's in the orbit of the
                # simple classical roots
                phi = -a0*L0(L.simple_roots()[0])
            else:
                phi = L0(L0.root_system.coroot_lattice().highest_root().associated_coroot())
            # Variant: try to fetch it from the other affinization; something like:
            # The a0 only has an influence in type BC; it handles the fact that alpha_0
            # is not in the orbit of the classical roots
            #phi1 = - L0(L'.other_affinization().simple_roots()[special_node]) * a0
            #assert phi == phi1

            j, v = phi.to_simple_root(reduced_word=True)
            translation = A0.monomial(-L0.simple_root(j)/a0)
            Tv = T[v]
            Tinv = T.Tw_inverse(v+(j,))

            def T0_check(weight):
                return -q1*q2*Tinv( translation * Tv(A0.monomial(weight)))
            # For debugging purposes
            T0_check.phi = phi
            T0_check.j = j
            T0_check.v = v
            return T0_check

        @cached_method
        def classical(self):
            """
            Return the group algebra of the corresponding classical lattice.

            EXAMPLES::

                sage: KL = RootSystem(["A",2,1]).ambient_space().algebra(QQ)
                sage: KL.classical()
                Algebra of the Ambient space of the Root system of type ['A', 2] over Rational Field
            """
            return self.basis().keys().classical().algebra(self.base_ring())

        def q_project_on_basis(self, l, q):
            r"""
            Return the monomial `c * cl(l)`  in the group algebra of the classical lattice.

            INPUT:

            - ``l`` -- an element of the root lattice realization
            - ``q`` -- an element of the ground ring

            Here, `cl(l)` is the projection of `l` in the classical
            lattice, and `c` is the coefficient of `l` in `\delta`.

            .. SEEALSO:: :meth:`q_project_on_basis`

            EXAMPLES::

                sage: K = QQ['q'].fraction_field()
                sage: q = K.gen()
                sage: KL = RootSystem(["A",2,1]).ambient_space().algebra(K)
                sage: L = KL.basis().keys()
                sage: e = L.basis()
                sage: KL.q_project_on_basis(4*e[1] + 3*e[2]
                ....:                        + e['deltacheck'] - 2*e['delta'], q)
                1/q^2*B[(0, 4, 3)]
            """
            KL0 = self.classical()
            L0 = KL0.basis().keys()
            return KL0.term(L0(l), q**l["delta"])

        def q_project(self, x, q):
            r"""
            Implement the `q`-projection morphism from ``self`` to the group algebra of the classical space.

            INPUT:

            - ``x`` -- an element of the group algebra of ``self``
            - ``q`` -- an element of the ground ring

            This is an algebra morphism mapping `\delta` to `q` and
            `X^b` to its classical counterpart for the other elements
            `b` of the basis of the realization.

            EXAMPLES::

                sage: K = QQ['q'].fraction_field()
                sage: q = K.gen()
                sage: KL = RootSystem(["A",2,1]).ambient_space().algebra(K)
                sage: L = KL.basis().keys()
                sage: e = L.basis()
                sage: x = (KL.an_element()
                ....:       + KL.monomial(4*e[1] + 3*e[2]
                ....:                      + e['deltacheck'] - 2*e['delta'])); x
                B[2*e[0] + 2*e[1] + 3*e[2]]
                 + B[4*e[1] + 3*e[2] - 2*e['delta'] + e['deltacheck']]
                sage: KL.q_project(x, q)
                B[(2, 2, 3)] + 1/q^2*B[(0, 4, 3)]

                sage: KL = RootSystem(["BC",3,2]).ambient_space().algebra(K)
                sage: L = KL.basis().keys()
                sage: e = L.basis()
                sage: x = (KL.an_element()
                ....:       + KL.monomial(4*e[1] + 3*e[2]
                ....:                      + e['deltacheck'] - 2*e['delta'])); x
                B[2*e[0] + 2*e[1] + 3*e[2]]
                 + B[4*e[1] + 3*e[2] - 2*e['delta'] + e['deltacheck']]
                sage: KL.q_project(x, q)
                B[(2, 2, 3)] + 1/q^2*B[(0, 4, 3)]

            .. WARNING::

                Recall that the null root, usually denoted `\delta`,
                is in fact ``a[0]\delta`` in Sage's notation, in order
                to avoid half integer coefficients (this only makes a
                difference in type BC). Similarly, what's usually
                denoted `q` is in fact ``q^a[0]`` in Sage's notations,
                to avoid manipulating square roots::

                    sage: KL.q_project(KL.monomial(L.null_root()),q)                    # needs sage.graphs
                    q^2*B[(0, 0, 0)]
            """
            L0 = self.classical()
            return L0.linear_combination( (self.q_project_on_basis(l, q), c) for l,c in x )

        def twisted_demazure_lusztig_operator_on_basis(self, weight, i, q1, q2, convention='antidominant'):
            r"""
            Return the twisted Demazure-Lusztig operator acting on the basis.

            INPUT:

            - ``weight`` -- an element `\lambda` of the weight lattice
            - ``i`` -- an element of the index set
            - ``q1``, ``q2`` -- two elements of the ground ring
            - ``convention`` -- ``'antidominant'``, ``'bar'``, or ``'dominant'`` (default: ``'antidominant'``)

            .. SEEALSO:: :meth:`twisted_demazure_lusztig_operators`

            EXAMPLES::

                sage: L = RootSystem(["A",3,1]).ambient_space()
                sage: e = L.basis()
                sage: K = QQ['q1,q2'].fraction_field()
                sage: q1, q2 = K.gens()
                sage: KL = L.algebra(K)
                sage: Lambda = L.classical().fundamental_weights()
                sage: KL.twisted_demazure_lusztig_operator_on_basis(
                ....:     Lambda[1] + 2*Lambda[2], 1, q1, q2, convention='dominant')
                -q2*B[(2, 3, 0, 0)]
                sage: KL.twisted_demazure_lusztig_operator_on_basis(
                ....:     Lambda[1] + 2*Lambda[2], 2, q1, q2, convention='dominant')
                -(q1+q2)*B[(3, 1, 1, 0)] - q2*B[(3, 0, 2, 0)]
                sage: KL.twisted_demazure_lusztig_operator_on_basis(
                ....:     Lambda[1] + 2*Lambda[2], 3, q1, q2, convention='dominant')
                q1*B[(3, 2, 0, 0)]
                sage: KL.twisted_demazure_lusztig_operator_on_basis(                    # needs sage.graphs
                ....:     Lambda[1]+2*Lambda[2], 0, q1, q2, convention='dominant')
                ((q1*q2+q2^2)/q1)*B[(1, 2, 1, 1)] + ((q1*q2+q2^2)/q1)*B[(1, 2, 2, 0)]
                 + q2^2/q1*B[(1, 2, 0, 2)] + ((q1^2+2*q1*q2+q2^2)/q1)*B[(2, 1, 1, 1)]
                 + ((q1^2+2*q1*q2+q2^2)/q1)*B[(2, 1, 2, 0)]
                 + ((q1*q2+q2^2)/q1)*B[(2, 1, 0, 2)]
                 + ((q1^2+2*q1*q2+q2^2)/q1)*B[(2, 2, 1, 0)]
                 + ((q1*q2+q2^2)/q1)*B[(2, 2, 0, 1)]
            """
            if i == 0: # should use the special node
                if convention != "dominant":
                    raise NotImplementedError("The twisted Demazure-Lusztig operator T_0 is only implemented in the dominant convention")
                return self.T0_check_on_basis(q1, q2, convention=convention)(weight)
            L = self.classical()
            return L.demazure_lusztig_operators(q1, q2, convention=convention)[i](L.monomial(weight))

        def twisted_demazure_lusztig_operators(self, q1, q2, convention='antidominant'):
            r"""
            Return the twisted Demazure-Lusztig operators acting on ``self``.

            INPUT:

            - ``q1``, ``q2`` -- two elements of the ground ring
            - ``convention`` -- ``'antidominant'``, ``'bar'``, or ``'dominant'`` (default: ``'antidominant'``)

            .. WARNING::

                - the code is currently only tested for `q_1q_2=-1`
                - only the ``'dominant'`` convention is functional for `i=0`

            For `T_1,\ldots,T_n`, these operators are the usual
            Demazure-Lusztig operators. On the other hand, the
            operator `T_0` is twisted::

                sage: L = RootSystem(["A",3,1]).ambient_space()
                sage: e = L.basis()
                sage: K = QQ['q1,q2'].fraction_field()
                sage: q1, q2 = K.gens()
                sage: KL = L.algebra(K)
                sage: T = KL.twisted_demazure_lusztig_operators(q1, q2, convention='dominant')
                sage: T._test_relations()

            TESTS:

            The following computations were checked with Mark Shimozono for type `A_1^{(1)}`::

                sage: L = RootSystem(["A",1,1]).ambient_space()
                sage: e = L.basis()
                sage: K = QQ['q1,q2'].fraction_field()
                sage: q1,q2 = K.gens()
                sage: KL = L.algebra(K)
                sage: T = KL.twisted_demazure_lusztig_operators(q1, q2, convention='dominant')
                sage: T._test_relations()
                sage: L0 = L.classical()
                sage: alpha = L0.simple_roots()
                sage: T.Ti_on_basis(L0.zero(), 1)
                q1*B[(0, 0)]
                sage: T.Ti_inverse_on_basis(L0.zero(), 1)
                1/q1*B[(0, 0)]
                sage: T.Ti_on_basis(alpha[1], 1)
                -(q1+q2)*B[(0, 0)] - q2*B[(-1, 1)]
                sage: T.Ti_inverse_on_basis(alpha[1], 1)
                ((q1+q2)/(q1*q2))*B[(0, 0)] + 1/q1*B[(-1, 1)] + ((q1+q2)/(q1*q2))*B[(1, -1)]
                sage: T.Ti_on_basis(L0.zero(), 0)                                       # needs sage.graphs
                (q1+q2)*B[(0, 0)] + q1*B[(1, -1)]

            The next computations were checked with Mark Shimozono for type `A_2^{(1)}`::

                sage: L = RootSystem(["A",2,1]).ambient_space()
                sage: e = L.basis()
                sage: K = QQ['u'].fraction_field()
                sage: u = K.gen()
                sage: KL = L.algebra(K)
                sage: T = KL.twisted_demazure_lusztig_operators(u, -~u, convention='dominant')
                sage: T._test_relations()
                sage: L0 = L.classical()
                sage: KL0 = L0.algebra(K)
                sage: alpha = L0.simple_roots()

                sage: phi = L0.highest_root(); phi
                (1, 0, -1)
                sage: phi.to_simple_root(reduced_word=True)
                (2, (1,))
                sage: res = T.Ti_on_basis(L0([1,0,1]), 1); res
                1/u*B[(0, 1, 1)]
                sage: res = res * KL0.monomial(-alpha[2]); res
                1/u*B[(0, 0, 2)]
                sage: res = T.Tw_inverse(2)(res); res
                ((u^2-1)/u^2)*B[(0, 1, 1)] + B[(0, 2, 0)]
                sage: res = T.Tw_inverse(1)(res); res
                ((u^2-1)/u)*B[(1, 1, 0)] + ((u^2-1)/u)*B[(1, 0, 1)] + u*B[(2, 0, 0)]

            .. TODO::

                Choose a good set of Cartan Type to run on. Rank >4 is
                too big. But `C_1` and `B_1` are boring.

            We now check systematically that those operators satisfy
            the relations of the Iwahori-Hecke algebra::

                sage: K = QQ['q1,q2'].fraction_field()
                sage: q1, q2 = K.gens()
                sage: for cartan_type in CartanType.samples(affine=True, crystallographic=True):  # long time (12s)
                ....:     if cartan_type.rank() > 4: continue
                ....:     if cartan_type.type() == 'BC': continue
                ....:     KL = RootSystem(cartan_type).weight_lattice().algebra(K)
                ....:     T = KL.twisted_demazure_lusztig_operators(q1, q2, convention='dominant')
                ....:     T._test_relations()

            .. TODO::

                Investigate why `T_0^\vee` currently does not satisfy
                the quadratic relation in type `BC`. This should
                hopefully be fixed when `T_0^\vee` will have a more
                uniform implementation::

                    sage: cartan_type = CartanType(["BC",1,2])
                    sage: KL = RootSystem(cartan_type).weight_lattice().algebra(K)
                    sage: T = KL.twisted_demazure_lusztig_operators(q1,q2, convention='dominant')
                    sage: T._test_relations()                                           # needs sage.graphs
                    Traceback (most recent call last):
                    ... tester.assertTrue(Ti(Ti(x,i,-q2),i,-q1).is_zero()) ...
                    AssertionError: False is not true

            Comparison with T0::

                sage: L = RootSystem(["A",2,1]).ambient_space()
                sage: e = L.basis()
                sage: K = QQ['t,q'].fraction_field()
                sage: t,q = K.gens()
                sage: q1 = t
                sage: q2 = -1
                sage: KL = L.algebra(K)
                sage: L0 = L.classical()
                sage: T = KL.demazure_lusztig_operators(q1,q2, convention='dominant')
                sage: def T0(*l0): return KL.q_project(T[0].on_basis()(L.embed_at_level(L0(l0), 1)), q)
                sage: T0_check_on_basis = KL.T0_check_on_basis(q1, q2,                  # needs sage.graphs
                ....:                                          convention='dominant')
                sage: def T0c(*l0): return T0_check_on_basis(L0(l0))

                sage: T0(0,0,1)                                 # not double checked    # needs sage.graphs
                -((t-1)/q)*B[(1, 0, 0)] + 1/q^2*B[(2, 0, -1)]
                sage: T0c(0,0,1)                                                        # needs sage.graphs
                (t^2-t)*B[(1, 0, 0)] + (t^2-t)*B[(1, 1, -1)] + t^2*B[(2, 0, -1)] + (t-1)*B[(0, 0, 1)]
            """
            T_on_basis = functools.partial(self.twisted_demazure_lusztig_operator_on_basis,
                                           q1=q1, q2=q2, convention=convention)
            return HeckeAlgebraRepresentation(self.classical(),
                                              T_on_basis,
                                              self.cartan_type().classical().dual().affine().dual(),
                                              q1, q2,
                                              side='left')

    class ElementMethods:

        def acted_upon(self, w):
            """
            Implement the action of ``w`` on ``self``.

            INPUT:

            - ``w`` -- an element of the Weyl group acting on the underlying weight lattice realization

            EXAMPLES::

                sage: L = RootSystem(["A",3]).ambient_space()
                sage: W = L.weyl_group()                                                # needs sage.libs.gap
                sage: M = L.algebra(QQ['q','t'])
                sage: m = M.an_element(); m  # TODO: investigate why we don't get something more interesting
                B[(2, 2, 3, 0)]
                sage: m = (m+1)^2; m
                B[(0, 0, 0, 0)] + 2*B[(2, 2, 3, 0)] + B[(4, 4, 6, 0)]
                sage: w = W.an_element(); w.reduced_word()                              # needs sage.libs.gap
                [1, 2, 3]
                sage: m.acted_upon(w)                                                   # needs sage.libs.gap
                B[(0, 0, 0, 0)] + 2*B[(0, 2, 2, 3)] + B[(0, 4, 4, 6)]
            """
            return self.map_support(w.action)

        def expand(self, alphabet):
            """
            Expand ``self`` into variables in the ``alphabet``.

            INPUT:

            - ``alphabet`` -- a non empty list/tuple of (invertible) variables in a ring to expand in

            EXAMPLES::

                sage: L = RootSystem(["A",2]).ambient_lattice()
                sage: KL = L.algebra(QQ)
                sage: p = KL.an_element() + KL.sum_of_monomials(L.some_elements()); p
                B[(1, 0, 0)] + B[(1, -1, 0)] + B[(1, 1, 0)] + 2*B[(2, 2, 3)] + B[(0, 1, -1)]
                sage: F = LaurentPolynomialRing(QQ, 'x,y,z')
                sage: p.expand(F.gens())
                2*x^2*y^2*z^3 + x*y + x + y*z^-1 + x*y^-1

            TESTS::

                sage: type(p.expand(F.gens()))
                <class 'sage.rings.polynomial.laurent_polynomial_mpair.LaurentPolynomial_mpair'>

                sage: p = KL.zero()
                sage: p.expand(F.gens())
                0
                sage: type(p.expand(F.gens()))
                <class 'sage.rings.polynomial.laurent_polynomial_mpair.LaurentPolynomial_mpair'>
            """
            codomain = alphabet[0].parent()
            return codomain.sum(c * prod(X**int(n)
                                         for X, n in zip(alphabet, vector(m)))
                                for m, c in self)



from sage.data_structures.bitset cimport *
from sage.data_structures.bitset_base cimport bitset_t, bitset_s
from sage.matroids.matroid cimport Matroid
from sage.matroids.set_system cimport SetSystem

cdef class BasisExchangeMatroid(Matroid):
    cdef long _groundset_size, _matroid_rank, _bitset_size
    cdef bitset_t _current_basis, _inside, _outside, _input, _input2, _output, _temp
    cdef tuple _E
    cdef dict _idx
    cdef frozenset _groundset

    cdef _bcount
    cdef Py_hash_t _weak_invariant_var, _strong_invariant_var, _heuristic_invariant_var
    cdef SetSystem _weak_partition_var, _strong_partition_var, _heuristic_partition_var

    cdef _relabel(self, mapping)

    cdef _pack(self, bitset_t, X)
    cdef __unpack(self, bitset_t)
    cdef bint _is_exchange_pair(self, long x, long y) except -1
    cdef int _exchange(self, long x, long y) except -1
    cdef int _move(self, bitset_t X, bitset_t Y) except -1
    cdef __fundamental_cocircuit(self, bitset_t, long x)
    cdef __fundamental_circuit(self, bitset_t, long y)

    cdef __max_independent(self, bitset_t, bitset_t)
    cdef __circuit(self, bitset_t, bitset_t)
    cdef __closure(self, bitset_t, bitset_t)
    cdef __max_coindependent(self, bitset_t, bitset_t)
    cdef __cocircuit(self, bitset_t, bitset_t)
    cdef _coclosure_internal(self, bitset_t, bitset_t)

    cdef __augment(self, bitset_t, bitset_t, bitset_t)
    cdef bint __is_independent(self, bitset_t F) except -1
    cdef __move_current_basis(self, bitset_t, bitset_t)

    cdef bint _set_current_basis(self, F) noexcept

    cpdef frozenset groundset(self)
    cpdef list groundset_list(self)
    cpdef full_rank(self)
    cpdef full_corank(self)

    cpdef basis(self)
    cpdef _move_current_basis(self, X, Y)

    cpdef frozenset _max_independent(self, frozenset F)
    cpdef int _rank(self, frozenset F) except? -1
    cpdef frozenset _circuit(self, frozenset F)
    cpdef frozenset _fundamental_circuit(self, frozenset B, e)
    cpdef frozenset _closure(self, frozenset F)

    cpdef frozenset _max_coindependent(self, frozenset F)
    cpdef int _corank(self, frozenset F) noexcept
    cpdef frozenset _cocircuit(self, frozenset F)
    cpdef frozenset _fundamental_cocircuit(self, frozenset B, e)
    cpdef frozenset _coclosure(self, frozenset F)

    cpdef frozenset _augment(self, frozenset X, frozenset Y)
    cpdef bint _is_independent(self, frozenset F) noexcept

    cpdef list whitney_numbers2(self)
    cdef  _whitney_numbers2_rec(self, object f_vec, bitset_t* flats, bitset_t* todo, long elt, long rnk)
    cdef  _flats_rec(self, SetSystem Rflats, long R, bitset_t* flats, bitset_t* todo, long elt, long rnk)
    cdef  _coflats_rec(self, SetSystem Rcoflats, long R, bitset_t* coflats, bitset_t* todo, long elt, long cornk)
    cdef _flat_element_inv(self, long k)
    cdef  _flat_element_inv_rec(self, object f_inc, long R, bitset_t* flats, bitset_t* todo, long elt, long i)

    cpdef bases_count(self)
    cpdef SetSystem independent_sets(self, long k=*)
    cpdef SetSystem dependent_sets(self, long k)

    cpdef SetSystem nonspanning_circuits(self)
    cpdef SetSystem cocircuits(self)
    cpdef SetSystem circuits(self, k=*)

    cpdef SetSystem _characteristic_setsystem(self)
    cpdef Py_hash_t _weak_invariant(self) noexcept
    cpdef SetSystem _weak_partition(self)
    cpdef Py_hash_t _strong_invariant(self) noexcept
    cpdef SetSystem _strong_partition(self)
    cpdef Py_hash_t _heuristic_invariant(self) noexcept
    cpdef SetSystem _heuristic_partition(self)
    cdef _flush_invariants(self)

    cpdef _equitable_partition(self, P=*)
    cpdef _is_isomorphic(self, other, certificate=*)
    cpdef _isomorphism(self, other)
    cpdef _is_isomorphism(self, other, morphism)
    cdef bint __is_isomorphism(self, BasisExchangeMatroid other, morphism) noexcept

    cpdef is_valid(self, certificate=*)

cdef bint nxksrd(bitset_s *b, long n, long k, bint succ) noexcept



def create_RealField(prec=53, type='MPFR', rnd='RNDN', sci_not=0):
    """
    Create a real field with given precision, type, rounding mode and
    scientific notation.

    Some options are ignored for certain types (RDF for example).

    INPUT:

    - ``prec`` -- positive integer

    - ``type`` -- type of real field:

      - ``'RDF'`` -- the Sage real field corresponding to native doubles
      - ``'Interval'`` -- real fields implementing interval arithmetic
      - ``'RLF'`` -- the real lazy field
      - ``'MPFR'`` -- floating point real numbers implemented using the MPFR
        library

    - ``rnd`` -- rounding mode:

      - ``'RNDN'`` -- round to nearest
      - ``'RNDZ'`` -- round toward zero
      - ``'RNDD'`` -- round down
      - ``'RNDU'`` -- round up

    - ``sci_not`` -- boolean, whether to use scientific notation for printing

    OUTPUT: the appropriate real field

    EXAMPLES::

        sage: from sage.rings.real_field import create_RealField
        sage: create_RealField(30)
        Real Field with 30 bits of precision
        sage: create_RealField(20, 'RDF')  # ignores precision
        Real Double Field
        sage: create_RealField(60, 'Interval')
        Real Interval Field with 60 bits of precision
        sage: create_RealField(40, 'RLF')  # ignores precision
        Real Lazy Field
    """
    if type == "RDF":
        from .real_double import RDF
        return RDF
    if type == "Interval":
        from .real_mpfi import RealIntervalField
        return RealIntervalField(prec, sci_not)
    if type == "Ball":
        from .real_arb import RealBallField
        return RealBallField(prec)
    if type == "RLF":
        from .real_lazy import RLF
        return RLF
    from .real_mpfr import RealField
    return RealField(prec, sci_not, rnd)



r"""
Space of morphisms of vector spaces (linear transformations)

AUTHOR:

    - Rob Beezer: (2011-06-29)

A :class:`VectorSpaceHomspace` object represents the set of all
possible homomorphisms from one vector space to another.
These mappings are usually known as linear transformations.

For more information on the use of linear transformations,
consult the documentation for vector space morphisms at
:mod:`sage.modules.vector_space_morphism`. Also, this is
an extremely thin veneer on free module homspaces
(:mod:`sage.modules.free_module_homspace`) and free module
morphisms (:mod:`sage.modules.free_module_morphism`) -
objects which might also be useful, and places
where much of the documentation resides.

EXAMPLES:

Creation and basic examination is simple. ::

    sage: V = QQ^3
    sage: W = QQ^2
    sage: H = Hom(V, W)
    sage: H
    Set of Morphisms (Linear Transformations) from
    Vector space of dimension 3 over Rational Field to
    Vector space of dimension 2 over Rational Field
    sage: H.domain()
    Vector space of dimension 3 over Rational Field
    sage: H.codomain()
    Vector space of dimension 2 over Rational Field

Homspaces have a few useful properties.  A basis is provided by
a list of matrix representations, where these matrix representatives
are relative to the bases of the domain and codomain.  ::

    sage: K = Hom(GF(3)^2, GF(3)^2)
    sage: B = K.basis()
    sage: for f in B:
    ....:     print(f)
    ....:     print("\n")
    Vector space morphism represented by the matrix:
    [1 0]
    [0 0]
    Domain: Vector space of dimension 2 over Finite Field of size 3
    Codomain: Vector space of dimension 2 over Finite Field of size 3
    <BLANKLINE>
    Vector space morphism represented by the matrix:
    [0 1]
    [0 0]
    Domain: Vector space of dimension 2 over Finite Field of size 3
    Codomain: Vector space of dimension 2 over Finite Field of size 3
    <BLANKLINE>
    Vector space morphism represented by the matrix:
    [0 0]
    [1 0]
    Domain: Vector space of dimension 2 over Finite Field of size 3
    Codomain: Vector space of dimension 2 over Finite Field of size 3
    <BLANKLINE>
    Vector space morphism represented by the matrix:
    [0 0]
    [0 1]
    Domain: Vector space of dimension 2 over Finite Field of size 3
    Codomain: Vector space of dimension 2 over Finite Field of size 3
    <BLANKLINE>

The zero and identity mappings are properties of the space.
The identity mapping will only be available if the domain and codomain
allow for endomorphisms (equal vector spaces with equal bases).  ::

    sage: H = Hom(QQ^3, QQ^3)
    sage: g = H.zero()
    sage: g([1, 1/2, -3])
    (0, 0, 0)
    sage: f = H.identity()
    sage: f([1, 1/2, -3])
    (1, 1/2, -3)

The homspace may be used with various representations of a
morphism in the space to create the morphism.  We demonstrate
three ways to create the same linear transformation between
two two-dimensional subspaces of ``QQ^3``.  The ``V.n`` notation
is a shortcut to the generators of each vector space, better
known as the basis elements.  Note that the matrix representations
are relative to the bases, which are purposely fixed when the
subspaces are created ("user bases").  ::

    sage: U = QQ^3
    sage: V = U.subspace_with_basis([U.0+U.1, U.1-U.2])
    sage: W = U.subspace_with_basis([U.0, U.1+U.2])
    sage: H = Hom(V, W)

First, with a matrix.  Note that the matrix representation
acts by matrix multiplication with the vector on the left.
The input to the linear transformation, ``(3, 1, 2)``,
is converted to the coordinate vector ``(3, -2)``, then
matrix multiplication yields the vector ``(-3, -2)``,
which represents the vector ``(-3, -2, -2)`` in the codomain.  ::

    sage: m = matrix(QQ, [[1, 2], [3, 4]])
    sage: f1 = H(m)
    sage: f1
    Vector space morphism represented by the matrix:
    [1 2]
    [3 4]
    Domain: Vector space of degree 3 and dimension 2 over Rational Field
    User basis matrix:
    [ 1  1  0]
    [ 0  1 -1]
    Codomain: Vector space of degree 3 and dimension 2 over Rational Field
    User basis matrix:
    [1 0 0]
    [0 1 1]
    sage: f1([3,1,2])
    (-3, -2, -2)

Second, with a list of images of the domain's basis elements.  ::

    sage: img = [1*(U.0) + 2*(U.1+U.2), 3*U.0 + 4*(U.1+U.2)]
    sage: f2 = H(img)
    sage: f2
    Vector space morphism represented by the matrix:
    [1 2]
    [3 4]
    Domain: Vector space of degree 3 and dimension 2 over Rational Field
    User basis matrix:
    [ 1  1  0]
    [ 0  1 -1]
    Codomain: Vector space of degree 3 and dimension 2 over Rational Field
    User basis matrix:
    [1 0 0]
    [0 1 1]
    sage: f2([3,1,2])
    (-3, -2, -2)

Third, with a linear function taking the domain to the codomain.  ::

    sage: g = lambda x: vector(QQ, [-2*x[0]+3*x[1], -2*x[0]+4*x[1], -2*x[0]+4*x[1]])
    sage: f3 = H(g)
    sage: f3
    Vector space morphism represented by the matrix:
    [1 2]
    [3 4]
    Domain: Vector space of degree 3 and dimension 2 over Rational Field
    User basis matrix:
    [ 1  1  0]
    [ 0  1 -1]
    Codomain: Vector space of degree 3 and dimension 2 over Rational Field
    User basis matrix:
    [1 0 0]
    [0 1 1]
    sage: f3([3,1,2])
    (-3, -2, -2)

The three linear transformations look the same, and are the same.  ::

    sage: f1 == f2
    True
    sage: f2 == f3
    True

TESTS::

    sage: V = QQ^2
    sage: W = QQ^3
    sage: H = Hom(QQ^2, QQ^3)
    sage: loads(dumps(H))
    Set of Morphisms (Linear Transformations) from
    Vector space of dimension 2 over Rational Field to
    Vector space of dimension 3 over Rational Field
    sage: loads(dumps(H)) == H
    True
"""

##############################################################################
#       Copyright (C) 2011 Rob Beezer <beezer@ups.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
##############################################################################

from sage.matrix.constructor import matrix
import sage.modules.free_module_homspace

# This module initially overrides just the minimum functionality necessary
# from  sage.modules.free_module_homspace.FreeModuleHomSpace.
# If additional methods here override the free module homspace methods,
# consider adjusting the free module doctests, since many are written with
# examples that are actually vector spaces and not so many use "pure" modules
# for the examples.


class VectorSpaceHomspace(sage.modules.free_module_homspace.FreeModuleHomspace):

    def __call__(self, A, check=True, **kwds):
        r"""
        INPUT:

        - ``A`` -- one of several possible inputs representing
          a morphism from this vector space homspace:

          - a vector space morphism in this homspace
          - a matrix representation relative to the bases of the vector spaces,
            which acts on a vector placed to the left of the matrix
          - a list or tuple containing images of the domain's basis vectors
          - a function from the domain to the codomain

        - ``check`` -- boolean (default: ``True``); required for compatibility
          with calls from :meth:`sage.structure.parent.Parent.hom`

        - the keyword ``side`` can be assigned the values ``'left'`` or
          ``'right'``; it corresponds to the side of vectors relative to the
          matrix

        EXAMPLES::

            sage: V = (QQ^3).span_of_basis([[1,1,0],[1,0,2]])
            sage: H = V.Hom(V)
            sage: H
            Set of Morphisms (Linear Transformations) from
            Vector space of degree 3 and dimension 2 over Rational Field
            User basis matrix:
            [1 1 0]
            [1 0 2]
            to
            Vector space of degree 3 and dimension 2 over Rational Field
            User basis matrix:
            [1 1 0]
            [1 0 2]

        Coercing a matrix::

            sage: A = matrix(QQ, [[0, 1], [1, 0]])
            sage: rho = H(A)          # indirect doctest
            sage: rho
            Vector space morphism represented by the matrix:
            [0 1]
            [1 0]
            Domain: Vector space of degree 3 and dimension 2 over Rational Field
            User basis matrix:
            [1 1 0]
            [1 0 2]
            Codomain: Vector space of degree 3 and dimension 2 over Rational Field
            User basis matrix:
            [1 1 0]
            [1 0 2]

        Coercing a list of images::

            sage: phi = H([V.1, V.0])
            sage: phi(V.1) == V.0
            True
            sage: phi(V.0) == V.1
            True
            sage: phi
            Vector space morphism represented by the matrix:
            [0 1]
            [1 0]
            Domain: Vector space of degree 3 and dimension 2 over Rational Field
            User basis matrix:
            [1 1 0]
            [1 0 2]
            Codomain: Vector space of degree 3 and dimension 2 over Rational Field
            User basis matrix:
            [1 1 0]
            [1 0 2]

        Coercing a lambda function::

            sage: f = lambda x: vector(QQ, [x[0], (1/2)*x[2], 2*x[1]])
            sage: zeta = H(f)
            sage: zeta
            Vector space morphism represented by the matrix:
            [0 1]
            [1 0]
            Domain: Vector space of degree 3 and dimension 2 over Rational Field
            User basis matrix:
            [1 1 0]
            [1 0 2]
            Codomain: Vector space of degree 3 and dimension 2 over Rational Field
            User basis matrix:
            [1 1 0]
            [1 0 2]

        Coercing a vector space morphism into the parent of a second vector
        space morphism will unify their parents::

            sage: U = FreeModule(QQ,3, sparse=True ); V = QQ^4
            sage: W = FreeModule(QQ,3, sparse=False); X = QQ^4
            sage: H = Hom(U, V)
            sage: K = Hom(W, X)
            sage: H is K, H == K
            (False, True)

            sage: A = matrix(QQ, 3, 4, [0]*12)
            sage: f = H(A)
            sage: B = matrix(QQ, 3, 4, range(12))
            sage: g = K(B)
            sage: f.parent() is H and g.parent() is K
            True

            sage: h = H(g)
            sage: f.parent() is h.parent()
            True

        See other examples in the module-level documentation.

        TESTS::

            sage: V = GF(3)^0
            sage: W = GF(3)^1
            sage: H = V.Hom(W)
            sage: H.zero().is_zero()
            True

        Previously the above code resulted in a :exc:`TypeError` because the
        dimensions of the matrix were incorrect.
        """
        from .vector_space_morphism import VectorSpaceMorphism
        D = self.domain()
        C = self.codomain()
        side = kwds.get("side", "left")
        from sage.structure.element import Matrix
        if isinstance(A, Matrix):
            pass
        elif isinstance(A, VectorSpaceMorphism):
            A = A.matrix()
        elif callable(A):
            try:
                images = [A(g) for g in D.basis()]
            except (ValueError, TypeError, IndexError) as e:
                msg = 'function cannot be applied properly to some basis element because\n' + e.args[0]
                raise ValueError(msg)
            try:
                A = matrix(D.dimension(), C.dimension(), [C.coordinates(C(a)) for a in images])
            except (ArithmeticError, TypeError) as e:
                msg = 'some image of the function is not in the codomain, because\n' + e.args[0]
                raise ArithmeticError(msg)
            if side == "right":
                A = A.transpose()
        elif isinstance(A, (list, tuple)):
            if len(A) != len(D.basis()):
                msg = "number of images should equal the size of the domain's basis (={0}), not {1}"
                raise ValueError(msg.format(len(D.basis()), len(A)))
            try:
                v = [C(a) for a in A]
                A = matrix(D.dimension(), C.dimension(), [C.coordinates(a) for a in v])
            except (ArithmeticError, TypeError) as e:
                msg = 'some proposed image is not in the codomain, because\n' + e.args[0]
                raise ArithmeticError(msg)
            if side == "right":
                A = A.transpose()
        else:
            msg = 'vector space homspace can only coerce matrices, vector space morphisms, functions or lists, not {0}'
            raise TypeError(msg.format(A))
        return VectorSpaceMorphism(self, A, side=side)

    def _repr_(self):
        r"""
        Text representation of a space of vector space morphisms.

        EXAMPLES::

            sage: H = Hom(QQ^2, QQ^3)
            sage: H._repr_().split(' ')
            ['Set', 'of', 'Morphisms', '(Linear', 'Transformations)',
            'from', 'Vector', 'space', 'of', 'dimension', '2', 'over',
            'Rational', 'Field', 'to', 'Vector', 'space', 'of',
            'dimension', '3', 'over', 'Rational', 'Field']
        """
        msg = 'Set of Morphisms (Linear Transformations) from {0} to {1}'
        return msg.format(self.domain(), self.codomain())



# distutils: libraries = flint
# distutils: depends = flint/nmod_poly_factor.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void nmod_poly_factor_init(nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_clear(nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_realloc(nmod_poly_factor_t fac, slong alloc) noexcept
    void nmod_poly_factor_fit_length(nmod_poly_factor_t fac, slong len) noexcept
    void nmod_poly_factor_set(nmod_poly_factor_t res, const nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_print(const nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_insert(nmod_poly_factor_t fac, const nmod_poly_t poly, slong exp) noexcept
    void nmod_poly_factor_concat(nmod_poly_factor_t res, const nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_pow(nmod_poly_factor_t fac, slong exp) noexcept
    bint nmod_poly_is_irreducible(const nmod_poly_t f) noexcept
    bint nmod_poly_is_irreducible_ddf(const nmod_poly_t f) noexcept
    bint nmod_poly_is_irreducible_rabin(const nmod_poly_t f) noexcept
    bint _nmod_poly_is_squarefree(mp_srcptr f, slong len, nmod_t mod) noexcept
    bint nmod_poly_is_squarefree(const nmod_poly_t f) noexcept
    void nmod_poly_factor_squarefree(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    bint nmod_poly_factor_equal_deg_prob(nmod_poly_t factor, flint_rand_t state, const nmod_poly_t pol, slong d) noexcept
    void nmod_poly_factor_equal_deg(nmod_poly_factor_t factors, const nmod_poly_t pol, slong d) noexcept
    void nmod_poly_factor_distinct_deg(nmod_poly_factor_t res, const nmod_poly_t poly, slong * const * degs) noexcept
    void nmod_poly_factor_distinct_deg_threaded(nmod_poly_factor_t res, const nmod_poly_t poly, slong * const * degs) noexcept
    void nmod_poly_factor_cantor_zassenhaus(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    void nmod_poly_factor_berlekamp(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    void nmod_poly_factor_kaltofen_shoup(nmod_poly_factor_t res, const nmod_poly_t poly) noexcept
    mp_limb_t nmod_poly_factor_with_berlekamp(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    mp_limb_t nmod_poly_factor_with_cantor_zassenhaus(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    mp_limb_t nmod_poly_factor_with_kaltofen_shoup(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    mp_limb_t nmod_poly_factor(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    void _nmod_poly_interval_poly_worker(void * arg_ptr) noexcept



"""
Center of a Universal Enveloping Algebra

AUTHORS:

- Travis Scrimshaw (2024-01-02): Initial version
"""

# ****************************************************************************
#       Copyright (C) 2024 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

# from sage.structure.unique_representation import UniqueRepresentation
# from sage.structure.parent import Parent
from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.integer_lists.invlex import IntegerListsLex
from sage.matrix.constructor import matrix
from sage.monoids.indexed_free_monoid import IndexedFreeAbelianMonoid
from sage.monoids.indexed_free_monoid import IndexedMonoid
from sage.combinat.root_system.coxeter_group import CoxeterGroup
from sage.combinat.integer_vector_weighted import iterator_fast as intvecwt_iterator
from sage.sets.non_negative_integers import NonNegativeIntegers
from sage.sets.finite_enumerated_set import FiniteEnumeratedSet
from sage.sets.family import Family
from sage.rings.integer_ring import ZZ
from sage.categories.kac_moody_algebras import KacMoodyAlgebras
from sage.categories.finite_dimensional_lie_algebras_with_basis import FiniteDimensionalLieAlgebrasWithBasis
from sage.categories.graded_algebras_with_basis import GradedAlgebrasWithBasis
from sage.categories.fields import Fields
from sage.categories.monoids import Monoids
from sage.categories.enumerated_sets import EnumeratedSets
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.data_structures.blas_dict import iaxpy
from collections import deque


class CenterIndices(IndexedFreeAbelianMonoid):
    r"""
    Set of basis indices for the center of a universal enveloping algebra.

    This also constructs the lift from the center to the universal enveloping
    algebra as part of computing the generators and basis elements. The
    basic algorithm is to construct the centralizer of each filtered
    component in increasing order (as each is a finite dimensional vector
    space). For more precise details, see [Motsak2006]_.
    """
    @staticmethod
    def __classcall__(cls, center):
        r"""
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: from sage.algebras.lie_algebras.center_uea import CenterIndices
            sage: g = lie_algebras.pwitt(GF(3), 3)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: CenterIndices(Z) is CenterIndices(Z)
            True
        """
        return super(IndexedMonoid, cls).__classcall__(cls, center)

    def __init__(self, center, indices=None):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: g = lie_algebras.pwitt(GF(5), 5)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: TestSuite(I).run(max_runs=7)
        """
        if indices is None:
            indices = NonNegativeIntegers()
        category = Monoids() & EnumeratedSets().Infinite()
        IndexedFreeAbelianMonoid.__init__(self, indices, prefix='Z', category=category)

        self._center = center
        self._envelop_alg = self._center._envelop_alg
        self._g = self._envelop_alg._g
        # The _lift_map will be a dict with the keys being the degree and the values
        #   given as dicts with the keys being the leading support. This will be
        #   used to do the corresponding reductions.
        self._lift_map = {0: {self._envelop_alg.one_basis(): self._envelop_alg.one()}}
        self._cur_deg = 0
        self._cur_vecs = deque()  # we do a lot of deletions in the middle
        # The _cur_basis is a mapping from the leading supports to a monoid element of self
        self._cur_basis = {self._envelop_alg.one_basis(): self.one()}
        self._cur_basis_inv = {self.one(): self._envelop_alg.one_basis()}
        self._gen_degrees = {}
        self._cur_num_gens = 0

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: g = lie_algebras.pwitt(GF(5), 5)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: Z.indices()
            Basis indices of Center of Universal enveloping algebra of
             The 5-Witt Lie algebra over Finite Field of size 5 in the Poincare-Birkhoff-Witt basis
        """
        return "Basis indices of {}".format(self._center)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: g = lie_algebras.pwitt(GF(5), 5)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: latex(I)
            B\left( Z\left( PBW\left( \mathcal{W}(5)_{\Bold{F}_{5}} \right) \right) \right)
        """
        from sage.misc.latex import latex
        return r"B\left( {} \right)".format(latex(self._center))

    def lift_on_basis(self, m):
        r"""
        Return the image of the basis element indexed by ``m`` in the
        universal enveloping algebra.

        EXAMPLES::

            sage: g = lie_algebras.Heisenberg(QQ, 3)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: z0 = I.monoid_generators()[0]
            sage: I._lift_map
            {0: {1: 1}}
            sage: I.lift_on_basis(z0)
            PBW['z']
            sage: I._lift_map
            {0: {1: 1}, 1: {PBW['z']: PBW['z']}}
            sage: I.lift_on_basis(z0^3)
            PBW['z']^3
            sage: I._lift_map
            {0: {1: 1}, 1: {PBW['z']: PBW['z']}}
            sage: I._construct_next_degree()
            sage: I._construct_next_degree()
            sage: I._lift_map
            {0: {1: 1},
             1: {PBW['z']: PBW['z']},
             2: {PBW['z']^2: PBW['z']^2},
             3: {PBW['z']^3: PBW['z']^3}}
            sage: I.lift_on_basis(z0^3)
            PBW['z']^3
        """
        while m not in self._cur_basis_inv:
            supp = m.support()
            # We might have not computed the correct degree, but we can lift the
            #   element if we have computed all of the corresponding generators.
            if all(i in self._gen_degrees and self._gen_degrees[i] in self._lift_map
                   for i in supp):
                ret = self._envelop_alg.one()
                divisors = [mp for mp in self._cur_basis_inv if mp.divides(m) and not mp.is_one()]
                while not m.is_one():
                    div = max(divisors, key=lambda elt: len(elt))
                    ls = self._cur_basis_inv[div]
                    deg = ls.length()
                    ret *= self._lift_map[deg][ls]
                    m = m // div
                    divisors = [mp for mp in divisors if mp.divides(m)]
                return ret
            self._construct_next_degree()
        ls = self._cur_basis_inv[m]
        deg = ls.length()
        return self._lift_map[deg][ls]

    def __iter__(self):
        r"""
        Iterate over ``self`` in degree increasing order.

        EXAMPLES::

            sage: g = lie_algebras.pwitt(GF(3), 6)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: it = iter(I)
            sage: [next(it) for _ in range(10)]
            [1, Z[0], Z[1], Z[2], Z[3], Z[4], Z[5], Z[6], Z[7], Z[0]^2]
        """
        yield self.one()  # start with the identity
        deg = 1
        while True:
            while deg not in self._lift_map:
                self._construct_next_degree()
            for ls in self._lift_map[deg]:
                yield self._cur_basis[ls]
            deg += 1

    def some_elements(self):
        r"""
        Return some elements of ``self``.

        EXAMPLES::

            sage: g = lie_algebras.pwitt(GF(3), 3)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: I.some_elements()
            [1, Z[0], Z[1], Z[2], Z[0]*Z[1]*Z[2], Z[0]*Z[2]^4, Z[0]^4*Z[1]^3]
        """
        it = iter(self)
        gens = [next(it) for _ in range(4)]
        # We construct it as a set in case we introduce duplicates.
        ret = set(gens)
        ret.update([self.prod(gens), gens[1] * gens[3]**4, gens[1]**4 * gens[2]**3])
        # Sort the output for uniqueness
        return sorted(ret, key=lambda m: (self.degree(m), m.to_word_list()))

    def degree(self, m):
        r"""
        Return the degre of ``m`` in ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: [I.degree(g) for g in I.monoid_generators()]
            [2, 5, 6, 8, 9, 12]
            sage: [(elt, I.degree(elt)) for elt in I.some_elements()]
            [(1, 0), (Z[0], 2), (Z[0]^2, 4), (Z[1], 5), (Z[0]^3*Z[1], 11),
             (Z[0]^10, 20), (Z[0]*Z[1]^4, 22)]
        """
        return ZZ.sum(e * self._gen_degrees[i] for i, e in m._monomial.items())

    def _construct_next_degree(self):
        r"""
        Construct the next elements of ``self`` for the next (uncomputed) degree.

        EXAMPLES::

            sage: g = lie_algebras.three_dimensional_by_rank(QQ, 2, 1)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: I._lift_map
            {0: {1: 1}}
            sage: I._construct_next_degree()
            sage: I._construct_next_degree()
            sage: I._construct_next_degree()
            sage: I._construct_next_degree()
            sage: I._lift_map
            {0: {1: 1}, 1: {}, 2: {}, 3: {}, 4: {}}
        """
        UEA = self._envelop_alg
        gens = UEA.algebra_generators()
        monoid = UEA.basis().keys()
        self._cur_deg += 1

        # We first update the lift map with all possible products
        # Note that we are using the fact that the elements are central
        #   so the product order doesn't matter.
        # Since we always update this, it is sufficient to compute it
        new_red = {}
        for i in range(1, self._cur_deg//2+1):
            for ls, lelt in self._lift_map[self._cur_deg-i].items():
                for rs, relt in self._lift_map[i].items():
                    supp = ls * rs
                    new_red[supp] = lelt * relt
                    mon = self._cur_basis[ls] * self._cur_basis[rs]
                    self._cur_basis[supp] = mon
                    self._cur_basis_inv[mon] = supp
        # TODO: Determine if we need to or benefit from another reduction of the new elements
        self._lift_map[self._cur_deg] = new_red

        # Determine the PBW elements of the current degree that are not reduced
        #   modulo the currently computed center.
        for exps in IntegerListsLex(n=self._cur_deg, length=len(gens)):
            elt = monoid.element_class(monoid, {k: p for k, p in zip(monoid._indices, exps) if p})
            if elt in new_red:  # already has a central element with this leading term
                continue
            # A new basis element to consider
            self._cur_vecs.append(UEA.monomial(elt))

        # Perform the centralization
        R = UEA.base_ring()
        vecs = list(self._cur_vecs)
        for g in gens:
            # TODO: We should hold onto previously computed values under the adjoint action
            ad = [g * v - v * g for v in vecs]
            # Compute the kernel
            supp = set()
            for v in ad:
                supp.update(v._monomial_coefficients)
            supp = sorted(supp, key=UEA._monomial_key, reverse=True)
            if not supp:  # no support for the image, so everything is in the kernel
                continue
            M = matrix(R, [[v[s] for v in ad] for s in supp])
            ker = M.right_kernel_matrix()
            vecs = [self._reduce(UEA.linear_combination((vecs[i], c) for i, c in kv.items()))
                    for kv in ker.rows()]

        # Lastly, update the appropriate data
        if not vecs: # No new central elements, so nothing to do
            return
        new_gens = {}
        for v in vecs:
            v = self._reduce(v)  # possibly not needed to check this
            if not v:
                continue
            ls = v.trailing_support(key=UEA._monomial_key)
            self._cur_vecs.remove(UEA.monomial(ls))
            new_gens[ls] = self._reduce(v)
            assert (self._cur_num_gens not in self._gen_degrees
                    or self._gen_degrees[self._cur_num_gens] == self._cur_deg)
            self._gen_degrees[self._cur_num_gens] = self._cur_deg
            mon = self.gen(self._cur_num_gens)
            self._cur_basis[ls] = mon
            self._cur_basis_inv[mon] = ls
            self._cur_num_gens += 1
        self._lift_map[self._cur_deg].update(new_gens)

    def _reduce(self, vec):
        r"""
        Return the UEA vector ``vec`` by the currently computed center.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: z0 = I.gen(0)
            sage: I._reduce(I.lift_on_basis(z0))
            0
            sage: max(I._lift_map)
            2
            sage: I._reduce(I.lift_on_basis(z0^2))
            4*PBW[alpha[1]]^2*PBW[-alpha[1]]^2
             + 2*PBW[alpha[1]]*PBW[alphacheck[1]]^2*PBW[-alpha[1]]
             + 1/4*PBW[alphacheck[1]]^4 - PBW[alphacheck[1]]^3
             + PBW[alphacheck[1]]^2
            sage: I._reduce(I.lift_on_basis(z0^2) - I.lift_on_basis(z0))
            4*PBW[alpha[1]]^2*PBW[-alpha[1]]^2
             + 2*PBW[alpha[1]]*PBW[alphacheck[1]]^2*PBW[-alpha[1]]
             + 1/4*PBW[alphacheck[1]]^4 - PBW[alphacheck[1]]^3
             + PBW[alphacheck[1]]^2
            sage: I._construct_next_degree()
            sage: I._construct_next_degree()
            sage: max(I._lift_map)
            4
            sage: I._reduce(I.lift_on_basis(z0^2) - I.lift_on_basis(z0))
            0
        """
        # This is replicating what SubmoduleWithBasis does
        ret = dict(vec._monomial_coefficients)
        for data in self._lift_map.values():
            for m, qv in data.items():
                if m not in ret:
                    continue
                iaxpy(-ret[m] / qv[m], qv._monomial_coefficients, ret)
        return self._envelop_alg._from_dict(ret, remove_zeros=False)


class SimpleLieCenterIndices(CenterIndices):
    r"""
    Set of basis indices for the center of a universal enveloping algebra of
    a simple Lie algebra.

    For more information, see
    :class:`~sage.algebras.lie_algebras.center_uea.CenterIndices`.
    """
    def __init__(self, center):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: TestSuite(I).run()
        """
        self._cartan_type = center._envelop_alg._g.cartan_type()
        r = self._cartan_type.rank()
        super().__init__(center, indices=FiniteEnumeratedSet(range(r)))
        W = CoxeterGroup(self._cartan_type)
        self._gen_degrees = dict(enumerate(W.degrees()))

    def __iter__(self):
        r"""
        Iterate over ``self`` in degree increasing order.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: it = iter(I)
            sage: [next(it) for _ in range(10)]
            [1, Z[0], Z[0]^2, Z[1], Z[2], Z[0]^3, Z[0]*Z[1], Z[3], Z[0]*Z[2], Z[0]^4]
        """
        deg = 0
        n = len(self._gen_degrees)
        wts = sorted(self._gen_degrees.values(), reverse=True)
        while True:
            for exps in intvecwt_iterator(deg, wts):
                yield self.element_class(self, {n-1-i: e for i, e in enumerate(exps) if e})
            deg += 1


class CenterUEA(CombinatorialFreeModule):
    r"""
    The center of a universal enveloping algebra.

    .. TODO::

        Generalize this to be the centralizer of any set of the UEA.

    .. TODO::

        For characteristic `p > 0`, implement the `p`-center of a simple
        Lie algebra. See, e.g.,

        - Theorem 5.12 of [Motsak2006]_
        - http://www.math.kobe-u.ac.jp/icms2006/icms2006-video/slides/059.pdf

    EXAMPLES::

        sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
        sage: U = g.pbw_basis()
        sage: Z = U.center()
        sage: B = Z.basis()
        sage: it = iter(B)
        sage: center_elts = [next(it) for _ in range(6)]; center_elts
        [1, Z[0], Z[1], Z[0]^2, Z[0]*Z[1], Z[1]^2]
        sage: elts = [U(v) for v in center_elts]  # long time
        sage: all(v * g == g * v for g in U.algebra_generators() for v in elts)  # long time
        True

    The Heisenberg Lie algebra `H_4` over a finite field; note the basis
    elements `b^p \in Z(U(H_4))` for the basis elements `b \in H_4`::

        sage: g = lie_algebras.Heisenberg(GF(3), 4)
        sage: U = g.pbw_basis()
        sage: Z = U.center()
        sage: B = Z.basis()
        sage: it = iter(B)
        sage: center_elts = [next(it) for _ in range(12)]; center_elts
        [1, Z[0], Z[0]^2, Z[0]^3, Z[1], Z[2], Z[3], Z[4], Z[5], Z[6], Z[7], Z[8]]
        sage: elts = [U(v) for v in center_elts]; elts
        [1, PBW['z'], PBW['z']^2, PBW['z']^3, PBW['p1']^3, PBW['p2']^3, PBW['p3']^3,
         PBW['p4']^3, PBW['q1']^3, PBW['q2']^3, PBW['q3']^3, PBW['q4']^3]
        sage: all(v * g == g * v for g in U.algebra_generators() for v in elts)
        True

    An example with a free 4-step nilpotent Lie algebras on 2 generators::

        sage: L = LieAlgebra(QQ, 2, step=4); L
        Free Nilpotent Lie algebra of rank 2 and step 4 over Rational Field
        sage: U = L.pbw_basis()
        sage: Z = U.center()
        sage: it = iter(Z.basis())
        sage: center_elts = [next(it) for _ in range(10)]; center_elts
        [1, Z[0], Z[1], Z[2], Z[0]^2, Z[0]*Z[1], Z[0]*Z[2], Z[1]^2, Z[1]*Z[2], Z[2]^2]
        sage: elts = [U(v) for v in center_elts]; elts
        [1, PBW[(1, 1, 1, 2)], PBW[(1, 1, 2, 2)], PBW[(1, 2, 2, 2)], PBW[(1, 1, 1, 2)]^2,
         PBW[(1, 1, 1, 2)]*PBW[(1, 1, 2, 2)], PBW[(1, 1, 1, 2)]*PBW[(1, 2, 2, 2)],
         PBW[(1, 1, 2, 2)]^2, PBW[(1, 1, 2, 2)]*PBW[(1, 2, 2, 2)], PBW[(1, 2, 2, 2)]^2]
        sage: all(v * g == g * v for g in U.algebra_generators() for v in elts)
        True

    Using the Engel Lie algebra::

        sage: L.<X,Y,Z> = LieAlgebra(QQ, {('X','Y'): {'Z': 1}}, nilpotent=True)
        sage: U = L.pbw_basis()
        sage: Z = U.center()
        sage: it = iter(Z.basis())
        sage: center_elts = [next(it) for _ in range(6)]; center_elts
        [1, Z[0], Z[0]^2, Z[0]^3, Z[0]^4, Z[0]^5]
        sage: elts = [U(v) for v in center_elts]; elts
        [1, PBW['Z'], PBW['Z']^2, PBW['Z']^3, PBW['Z']^4, PBW['Z']^5]
        sage: all(v * g == g * v for g in U.algebra_generators() for v in elts)
        True
    """
    def __init__(self, g, UEA):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(ZZ['t'].fraction_field(), cartan_type=['D', 4])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: TestSuite(Z).run()

            sage: g = lie_algebras.Heisenberg(GF(3), 4)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: TestSuite(Z).run()
        """
        if g not in FiniteDimensionalLieAlgebrasWithBasis:
            raise NotImplementedError("only implemented for finite dimensional Lie algebras with a distinguished basis")

        R = UEA.base_ring()
        if R not in Fields():
            raise NotImplementedError("only implemented for the base ring a field")

        self._g = g
        self._envelop_alg = UEA
        if (self._g in KacMoodyAlgebras
            and self._g.cartan_type().is_finite()
            and R.characteristic() == 0):
            indices = SimpleLieCenterIndices(self)
        else:
            indices = CenterIndices(self)
        category = UEA.category()
        base = category.base()
        category = GradedAlgebrasWithBasis(base).Commutative() | category.Subobjects()
        CombinatorialFreeModule.__init__(self, R, indices, category=category,
                                         prefix='', bracket=False, latex_bracket=False,
                                         sorting_key=self._sorting_key)
        self.lift.register_as_coercion()

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A',2])
            sage: U = g.pbw_basis()
            sage: U.center()
            Center of Universal enveloping algebra of Lie algebra of ['A', 2]
             in the Chevalley basis in the Poincare-Birkhoff-Witt basis
        """
        return "Center of " + repr(self._envelop_alg)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: g = lie_algebras.pwitt(GF(5), 5)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: latex(Z)
            Z\left( PBW\left( \mathcal{W}(5)_{\Bold{F}_{5}} \right) \right)
        """
        from sage.misc.latex import latex
        return r"Z\left( {} \right)".format(latex(self._envelop_alg))

    def _sorting_key(self, m):
        r"""
        Return a key for ``m`` used in sorting elements of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: z0, z1 = Z.algebra_generators()
            sage: z0 * z1  # indirect doctest
            Z[0]*Z[1]
            sage: z1^2 + z0*z1 + z0^2  # indirect doctest
            Z[0]^2 + Z[0]*Z[1] + Z[1]^2
            sage: z1^3 + z0*z1^2 + z0^2*z1 + z0^3  # indirect doctest
            Z[0]^3 + Z[0]^2*Z[1] + Z[0]*Z[1]^2 + Z[1]^3
        """
        return (-m.length(), m.to_word_list())

    @cached_method
    def algebra_generators(self):
        r"""
        Return the algebra generators of ``self``.

        .. WARNING::

            When the universal enveloping algebra is not known to have
            a finite generating set, the generating set will be the basis
            of ``self`` in a degree (weakly) increasing order indexed by
            `\ZZ_{\geq 0}`. In particular, the `0`-th generator will be
            the multiplicative identity `1`.

        EXAMPLES::

            sage: g = lie_algebras.Heisenberg(QQ, 3)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: Z.algebra_generators()[0]
            1
            sage: Z.algebra_generators()[1]
            Z[0]

            sage: g = LieAlgebra(QQ, cartan_type=['G', 2])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: Z.algebra_generators()
            Finite family {0: Z[0], 1: Z[1]}
        """
        mon_gens = self._indices.monoid_generators()
        if mon_gens.cardinality() == float("inf"):
            return Family(NonNegativeIntegers(), lambda m: self.monomial(self._indices.unrank(m)))
        return Family({i: self.monomial(mon_gens[i]) for i in mon_gens.keys()})

    @cached_method
    def one_basis(self):
        r"""
        Return the basis index of `1` in ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ['t'].fraction_field(), cartan_type=['B', 5])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: ob = Z.one_basis(); ob
            1
            sage: ob.parent() is Z.indices()
            True
        """
        return self._indices.one()

    def ambient(self):
        r"""
        Return the ambient algebra of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(GF(5), cartan_type=['A', 2])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: Z.ambient() is U
            True
        """
        return self._envelop_alg

    def product_on_basis(self, left, right):
        r"""
        Return the product of basis elements indexed by ``left`` and ``right``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: mg = Z.indices().monoid_generators()
            sage: Z.product_on_basis(mg[1]*mg[2], mg[0]*mg[1]^3*mg[2]*mg[3]^3)
            Z[0]*Z[1]^4*Z[2]^2*Z[3]^3
        """
        return self.monomial(left * right)

    def degree_on_basis(self, m):
        r"""
        Return the degree of the basis element indexed by ``m`` in ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: I = Z.indices()
            sage: it = iter(I)
            sage: supports = [next(it) for _ in range(10)]; supports
            [1, Z[0], Z[0]^2, Z[1], Z[2], Z[0]^3, Z[0]*Z[1], Z[3], Z[0]*Z[2], Z[0]^4]
            sage: [Z.degree_on_basis(m) for m in supports]
            [0, 2, 4, 5, 6, 6, 7, 8, 8, 8]
        """
        return self._indices.degree(m)

    @lazy_attribute
    def lift(self):
        r"""
        The lift map from ``self`` to the universal enveloping algebra.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: gens = Z.algebra_generators()
            sage: U(gens[0]^2 + gens[0])
            4*PBW[alpha[1]]^2*PBW[-alpha[1]]^2
             + 2*PBW[alpha[1]]*PBW[alphacheck[1]]^2*PBW[-alpha[1]]
             + 1/4*PBW[alphacheck[1]]^4 - PBW[alphacheck[1]]^3
             - 2*PBW[alpha[1]]*PBW[-alpha[1]] + 1/2*PBW[alphacheck[1]]^2
             + PBW[alphacheck[1]]
            sage: U(-1/4*gens[0]) == U.casimir_element()
            True
        """
        # This is correct if we are using key=self._envelop_alg._monomial_key,
        #   but we are currently unable to pass such an option.
        return self.module_morphism(self._indices.lift_on_basis, codomain=self._envelop_alg, unitriangular='upper')

    def retract(self, elt):
        r"""
        The retraction map to ``self`` from the universal enveloping algebra.

        .. TODO::

            Implement a version of this that checks if the leading term of
            ``elt`` is divisible by a product of all of the currently known
            generators in order to avoid constructing the full centralizer
            of larger degrees than needed.

        EXAMPLES::

            sage: g = lie_algebras.Heisenberg(QQ, 3)
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: z0 = Z.algebra_generators()[1]; z0
            Z[0]
            sage: Z.retract(U(z0^2) - U(3*z0))
            Z[0]^2 - 3*Z[0]

            sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
            sage: U = g.pbw_basis()
            sage: Z = U.center()
            sage: z0, z1 = Z.algebra_generators()
            sage: Z.retract(U(z0*z0) - U(z1))  # long time
            Z[0]^2 - Z[1]
            sage: zc = Z.retract(U.casimir_element()); zc
            -1/3*Z[0]
            sage: U(zc) == U.casimir_element()
            True
        """
        # This should work except it needs the monomials of the PBW basis to be
        # comparable. However, this does not work for, e.g., Lie algebras
        # in the Chevalley basis as ee are unable to pass a key for the
        # module morphism. Additionally, the implementation below does more
        # operations in-place than the module morphism.
        #return self.lift.section()
        UEA = self._envelop_alg
        elt = UEA(elt)
        # We manipulate the dictionary (in place) to avoid creating elements
        data = elt.monomial_coefficients(copy=True)
        indices = self._indices
        ret = {}
        while data:
            lm = min(data, key=UEA._monomial_key)
            while indices._cur_deg < UEA.degree_on_basis(lm):
                indices._construct_next_degree()
            ind = indices._cur_basis[lm]
            other = indices.lift_on_basis(ind).monomial_coefficients(copy=False)
            coeff = data[lm] / other[lm]
            ret[ind] = coeff
            iaxpy(-coeff, other, data)
        return self.element_class(self, ret)



import pytest
from sage.rings.padics.factory import ZpLC, ZpLF, QpLC, QpLF

# ZpLC, ZpLF, QpLC, and QpLF all raise FutureWarnings
from warnings import catch_warnings, filterwarnings
filterwarnings("ignore", category=FutureWarning)


@pytest.fixture
def R1():
    return ZpLC(2)


@pytest.fixture
def R2():
    return ZpLF(2)


@pytest.fixture
def R3():
    return QpLC(2)


@pytest.fixture
def R4():
    return QpLF(2)


# Use strings for the fixture names here, and then later convert them
# to the actual fixture objects using request.getfixturevalue(). This
# is a workaround for being unable to pass fixtures directly as
# parameters:
#
#   https://github.com/pytest-dev/pytest/issues/349
#
elements = ( "R1", "R2", "R3", "R4" )


@pytest.mark.parametrize("e", elements)
def test_padic_lattice_element(e, request):
    r"""
    Run the ``TestSuite()`` for some examples that previously
    lived in the TESTS:: block of the padic_lattice_element module.
    """
    from sage.misc.sage_unittest import TestSuite

    # Convert the string to a real fixture
    e = request.getfixturevalue(e)

    # Only do a few runs, _test_matrix_smith() in particular is slow.
    TestSuite(e).run(verbose=True,
                     raise_on_failure=True,
                     skip="_test_teichmuller",
                     max_runs=8)



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_sf_exp.h":

  double  gsl_sf_exp(double x)

  int  gsl_sf_exp_e(double x, gsl_sf_result * result)

  int  gsl_sf_exp_e10_e(double x, gsl_sf_result_e10 * result)

  double  gsl_sf_exp_mult(double x, double y)

  int  gsl_sf_exp_mult_e(double x, double y, gsl_sf_result * result)

  int  gsl_sf_exp_mult_e10_e(double x, double y, gsl_sf_result_e10 * result)

  double  gsl_sf_expm1(double x)

  int  gsl_sf_expm1_e(double x, gsl_sf_result * result)

  double  gsl_sf_exprel(double x)

  int  gsl_sf_exprel_e(double x, gsl_sf_result * result)

  double  gsl_sf_exprel_2(double x)

  int  gsl_sf_exprel_2_e(double x, gsl_sf_result * result)

  double  gsl_sf_exprel_n(int n, double x)

  int  gsl_sf_exprel_n_e(int n, double x, gsl_sf_result * result)

  int  gsl_sf_exp_err_e(double x, double dx, gsl_sf_result * result)

  int  gsl_sf_exp_err_e10_e(double x, double dx, gsl_sf_result_e10 * result)

  int  gsl_sf_exp_mult_err_e(double x, double dx, double y, double dy, gsl_sf_result * result)

  int  gsl_sf_exp_mult_err_e10_e(double x, double dx, double y, double dy, gsl_sf_result_e10 * result)

